import logging

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks

from vortex.DeferUtil import vortexLogFailure
from vortex.VortexFactory import VortexFactory
from vortex.rpc.test.RPCTest import myRemoteAddMethod

logger = logging.getLogger(__name__)
logging.basicConfig()


@inlineCallbacks
def connect():
    yield VortexFactory.createWebsocketClient(
        "sendVortexName", "127.0.0.1", 10101, "ws://127.0.0.1:10101/vortexws"
    )

    reactor.callLater(2, callRpc)


def callRpc():
    d = myRemoteAddMethod(7, kwarg1=4)
    d.addCallback(lambda v: logger.debug("SUCCESSS, result = %s", v))
    d.addErrback(vortexLogFailure, logger)
    d.addBoth(lambda _: reactor.stop())
    return d


if __name__ == "__main__":
    reactor.callLater(0, logger.info, "RPCTest client running")
    reactor.callLater(0, connect)
    reactor.run()



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import inspect
import logging
import types
import weakref
from copy import copy
from typing import Callable, Optional, Dict, Union

from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.PayloadIO import PayloadIO
from vortex.VortexABC import SendVortexMsgResponseCallable

logger = logging.getLogger(__name__)


# TODO: typings needs to support keyword args before tbis will work.
# PayloadEndpointProcessCallable = Callable[
#     [
#         ...
#         # PayloadEnvelope,  # payloadEnvelope
#         # str,  # vortexUuid
#         # str,  # vortexName
#         # Any,  # httpSession
#         # SendVortexMsgResponseCallable  # sendResponse
#     ],
#     Optional[Deferred]  # Return
# ]


class PayloadEndpoint(object):
    """
    The payload endpoint is responsible for matching payloads filters against
    filters defined in the endpoint. If the end point filters are within the
    payload filter then the payload will be passed to the supplied callable.
    """

    PERMITTER = None

    def __init__(
        self,
        filt: Dict,
        callable_,
        acceptOnlyFromVortex: Optional[Union[str, tuple]] = None,
    ) -> None:
        """
        :param filt: The filter to match against payloads
        :param callable_: This will be called and passed the payload if it matches
        :param acceptOnlyFromVortex: Accept payloads only from this vortex,
            The vortex can be str or tuple of str, or None to accept from any.
        """
        self._acceptOnlyFromVortex = acceptOnlyFromVortex
        if isinstance(self._acceptOnlyFromVortex, str):
            self._acceptOnlyFromVortex = (self._acceptOnlyFromVortex,)

        if not "key" in filt:
            e = Exception(
                "There is no 'key' in the payload filt"
                ", There must be one for routing"
            )
            logger.exception(e)
            raise e

        self._wref: Callable[[], Optional[Callable]] = None
        if isinstance(callable_, types.FunctionType):
            w = None
            if hasattr(callable_, "_endpointWeakClass"):
                w = callable_._endpointWeakClass

            else:

                class W:
                    def __init__(self, callable_):
                        self._callable = callable_
                        self._callable._endpointWeakClass = self

                    def __call__(self, payload):
                        self._callable(payload)

                w = W(callable_)

            self._wref = weakref.ref(w)

        else:
            weakObject = weakref.ref(callable_.__self__)
            weakMethod = weakref.ref(callable_.__func__)

            if callable_.__func__.__name__ == "func":
                logger.warning(
                    "The registered callback is 'func' this is likely"
                    " a method wrapped in @deferToThreadWrapWithLogger, %s",
                    filt,
                )

            def getCallable():
                obj = weakObject()
                func = weakMethod()
                if obj and func:
                    return getattr(obj, func.__name__)
                return None

            self._wref = getCallable

        self._filt = filt
        PayloadIO().add(self)

    @property
    def filt(self):
        return copy(self._filt)

    def check(self, payloadEnvelope: PayloadEnvelope, vortexName: str) -> bool:

        # Filter for the backend vortexes.
        if (
            self._acceptOnlyFromVortex
            and vortexName not in self._acceptOnlyFromVortex
        ):
            return False

        def removeUnhashable(filt):
            items = set()
            for key, value in list(filt.items()):
                # We can compare an array of primitives
                if isinstance(value, list):
                    value = tuple(sorted(value))

                # We don't compare complex structures
                if isinstance(value, dict):
                    continue

                if inspect.isclass(value) and (
                    issubclass(value, dict) or issubclass(value, list)
                ):
                    raise Exception(
                        "Class type passed instead of an instance"
                        " key:%s, value:%s" % (key, value)
                    )

                items.add((key, value))
            return items

        theirFilt = removeUnhashable(payloadEnvelope.filt)
        ourFilt = removeUnhashable(self._filt)

        return set(ourFilt).issubset(theirFilt)

    def process(
        self,
        payloadEnvelope: PayloadEnvelope,
        vortexUuid: str,
        vortexName: str,
        httpSession,
        sendResponse: SendVortexMsgResponseCallable,
    ):

        if not self.check(payloadEnvelope, vortexName):
            return

        if self.PERMITTER and not self.PERMITTER.check(
            payloadEnvelope, httpSession
        ):
            logger.debug("Permission failed for %s", payloadEnvelope.filt)
            return

        callable_ = self._wref()
        if callable_:
            return callable_(
                payloadEnvelope=payloadEnvelope,
                vortexUuid=vortexUuid,
                vortexName=vortexName,
                httpSession=httpSession,
                sendResponse=sendResponse,
            )
        else:
            PayloadIO().remove(self)

    def _callableExpired(self, expiredCallable):
        pass
        # PayloadIO().remove(self)

    def __repr__(self):
        callable_ = self._wref()
        if callable_:
            try:
                callbackStr = "%s.%s" % (
                    callable_.__self__.__class__.__name__,
                    callable_.__name__,
                )
            except Exception as e:
                callbackStr = str(e)
        else:
            callbackStr = "None"

        s = "Payload id=%s\nfilt=%s\ncallback=%s"
        return s % (id(self), self._filt, callbackStr)

    def shutdown(self):
        PayloadIO().remove(self)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import inspect
import json
import re
from abc import ABCMeta, abstractmethod, ABC
from copy import deepcopy
from datetime import datetime
from time import strptime
from typing import Dict, Optional, Annotated, Any
from typing import List
from typing import Type
from typing import get_args
from typing import get_origin
from typing import get_type_hints

from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.properties import RelationshipProperty
from sqlalchemy.sql.schema import Sequence

from vortex.SerialiseUtil import ISO8601_REGEXP
from .Jsonable import Jsonable
from .SerialiseUtil import ISO8601
from .SerialiseUtil import T_RAPUI_TUPLE
from .SerialiseUtil import WKBElement
from .SerialiseUtil import convertFromWkbElement

TUPLE_TYPES = []
TUPLE_TYPES_BY_NAME: Dict[str, "Tuple"] = {}
TUPLE_TYPES_BY_SHORT_NAME: Dict[str, "Tuple"] = {}

JSON_EXCLUDE = "jsonExclude"


def registeredTupleNames():
    return list(TUPLE_TYPES_BY_NAME.keys())


def tupleForTupleName(tupleName):
    return TUPLE_TYPES_BY_NAME[tupleName]


def getSqlaRelationshipFieldNames(cls):
    objDict = _getTupleInternalDict(cls)

    # Tell the tuple fields of their variable name
    fields = []
    for name, value in objDict.items():
        if isinstance(value, InstrumentedAttribute) and isinstance(
            value.comparator.prop, RelationshipProperty
        ):
            fields.append(value)

    return fields


def addTupleType(cls):
    if cls.tupleName() is None:
        raise Exception(
            "You must set __tupleType__ to a unique string for this tuple"
        )

    if not cls.tupleName():
        cls.__tupleType__ = "%s.%s" % (cls.__module__, cls.__name__)

    tupleType = cls.tupleName()
    tupleTypeShort = cls.__tupleTypeShort__
    if tupleType in TUPLE_TYPES_BY_NAME or tupleType is None:
        raise Exception(
            "Tuple name is None or is already in registered.\n"
            "Tuple name is %s" % tupleType
        )

    if tupleTypeShort:
        if tupleTypeShort in TUPLE_TYPES_BY_SHORT_NAME:
            raise Exception(
                "Tuple short name is already registered.\n"
                "Tuple short name is %s" % tupleTypeShort
            )

        TUPLE_TYPES_BY_SHORT_NAME[tupleTypeShort] = cls

    # Setup the lookups
    TUPLE_TYPES.append(cls)
    TUPLE_TYPES_BY_NAME[tupleType] = cls

    def underscoreException():
        raise Exception(
            "TupleFields can not start with an underscore. "
            "They can potentially clash with inner tuple workings"
            " if they do."
        )

    # This is required because SQLAlchemy seems to alter the inhertance.
    # OR __dict__ just doesn't contian attributes defined in the base classes
    # If we inherit from a tuple that has a tuple field, that field should be included
    def getBaseClassTupleFields(cls, objDict):
        # Skip these types
        if cls in (object, Tuple, Jsonable):
            return

        for name, value in cls.__dict__.items():
            # SQLAlchemy takes care of it's own fields.
            if isinstance(value, TupleField):
                objDict[name] = value

    objDict = {}
    for baseCls in reversed(inspect.getmro(cls)[1:]):  # The first item is us
        getBaseClassTupleFields(baseCls, objDict)
    objDict.update(cls.__dict__)

    # Tell the tuple fields of their variable name
    fields = []
    shortFieldNames = set()
    for name, value in objDict.items():

        if isinstance(value, TupleField):
            shortName = value.shortName if value.shortName else name

        elif isinstance(value, InstrumentedAttribute):
            if isinstance(value.comparator.prop, RelationshipProperty):
                continue

            shortName = value.doc if value.doc else name

        else:
            continue

        if name.startswith("_"):
            underscoreException()

        if shortName.startswith("_"):
            underscoreException()

        value.name = name
        fields.append(name)

        if shortName != JSON_EXCLUDE:
            if shortName in shortFieldNames:
                raise Exception(
                    "TupleField %s short name %s is alread registered"
                    % (name, shortName)
                )

            shortFieldNames.add(shortName)

    hasSlots = hasattr(cls, "__slots__")
    hasFieldNames = cls.__fieldNames__ is not None

    if hasSlots and hasFieldNames:
        raise Exception(
            "Only one of __slots__ or __fieldNames__ can be defined"
            " but not both"
        )

    # If field names already exist, then work off these
    if hasFieldNames:
        fields = list(set(fields + list(cls.__fieldNames__)))

        # Just check that the field names are defined.
        for fieldName in fields:
            if not hasattr(cls, fieldName):
                raise Exception(
                    "Tuple %s doesn't have field %s" % (tupleType, fieldName)
                )

    # If field names already exist, then work off these
    if hasSlots:
        fields = list(set(fields + list(cls.__slots__)))

    # Sort and set the field names
    fields.sort()
    cls.__fieldNames__ = fields

    __mapShortFieldNames(cls)

    return cls


def _getTupleInternalDict(cls):
    """Get Tuple Internal Dict

    This is required because SQLAlchemy seems to alter the inhertance.
    OR __dict__ just doesn't contian attributes defined in the base classes
    If we inherit from a tuple that has a tuple field, that field should be included
    """

    def getBaseClassTupleFields(cls, objDict):
        # Skip these types
        if cls in (object, Tuple, Jsonable):
            return

        for name, value in cls.__dict__.items():
            # SQLAlchemy takes care of it's own fields.
            if isinstance(value, TupleField):
                objDict[name] = value

    objDict = {}
    for baseCls in reversed(inspect.getmro(cls)[1:]):  # The first item is us
        getBaseClassTupleFields(baseCls, objDict)
    objDict.update(cls.__dict__)
    return objDict


def __mapShortFieldNames(cls):
    fieldMap = {}
    for normalName in cls.__fieldNames__:
        field = getattr(cls, normalName, None)
        if not field:
            continue

        if isinstance(field, TupleField):
            if field.jsonExclude:
                continue

            shortName = field.shortName if field.shortName else field.name

        elif isinstance(field, InstrumentedAttribute):
            if isinstance(field.comparator.prop, RelationshipProperty):
                continue

            shortName = field.doc if field.doc else field.name

        else:
            continue

        # Underscore means skip
        if shortName == JSON_EXCLUDE:
            continue

        fieldMap[shortName] = normalName

    cls.__shortFieldNamesMap__ = fieldMap


def removeTuplesForPackage(packageName):
    tupleNames = [
        cls.tupleName()
        for cls in TUPLE_TYPES
        if cls.__name__.startswith("%s." % packageName)
    ]

    removeTuplesForTupleNames(tupleNames)


def removeTuplesForTupleNames(tupleNames):
    global TUPLE_TYPES, TUPLE_TYPES_BY_NAME, TUPLE_TYPES_BY_SHORT_NAME

    tupleNames = set(tupleNames)

    def filt(cls):
        return cls.tupleName() not in tupleNames

    tupleShortNames = [
        cls.__tupleTypeShort__ for cls in TUPLE_TYPES if filt(cls)
    ]

    # Remove from tuple types
    TUPLE_TYPES = list(filter(filt, TUPLE_TYPES))

    # Remove from tuple types by name
    for tupleName in tupleNames:
        if tupleName in TUPLE_TYPES_BY_NAME:
            del TUPLE_TYPES_BY_NAME[tupleName]

    # Remove from tuple types by name
    for tupleShortName in tupleShortNames:
        if tupleShortName in TUPLE_TYPES_BY_SHORT_NAME:
            del TUPLE_TYPES_BY_SHORT_NAME[tupleShortName]


class TupleField:
    class _Map:
        pass

    def __init__(
        self,
        defaultValue=None,
        typingType=None,
        comment="",
        shortName=None,
        jsonExclude=False,
    ):
        self.name = None
        self.shortName = shortName
        self.defaultValue = defaultValue
        self.typingType = typingType
        self.comment = comment
        self.jsonExclude = jsonExclude


class PolymorphicTupleTypeFieldArg:
    def __init__(self, fieldName: str):
        self.fieldName = fieldName


class _TupleToSqlaJsonMixin:
    def tupleToSqlaBulkInsertDict(self, includeNulls=True):
        insertDict = {}

        for field in self.__class__.__dict__.values():

            if not isinstance(field, InstrumentedAttribute):
                continue

            if isinstance(field.comparator.prop, RelationshipProperty):
                continue

            def convert(value):
                if isinstance(value, WKBElement):
                    return value.desc

                else:
                    return value

            value = getattr(self, field.name)
            if includeNulls or value is not None:
                insertDict[field.name] = convert(value)

        return insertDict


class TupleFieldValidatorABC(ABC):
    """Tuple Value Validator

    This class can be used to validate value ranges within a vortex.Tuple field

    """

    @abstractmethod
    def validate(self, fieldName: str, value):
        """Validate

        This method is called to validate the value.
        If the value is invalid a ValueError will be raised.

        :param fieldName: The name of the field being validated
        :param value: The value to be checked
        """


class IntTupleFieldValidator(TupleFieldValidatorABC):
    def __init__(self, low: Optional[int], high: Optional[int]):
        self._low = low
        self._high = high

    def validate(self, fieldName: str, value):
        if self._low is not None and value < self._low:
            raise ValueError(
                f"Field {fieldName}," f" Value {value} is less than {self._low}"
            )

        if self._high is not None and self._high < value:
            raise ValueError(
                f"Field {fieldName},"
                f" Value {value} is greater than {self._high}"
            )


def getConcreteClass(TupleClass, jsonDict):
    # Create concrete class for Tuple
    if not hasattr(TupleClass, "__tupleArgs__"):
        return TupleClass

    polyType = [
        p
        for p in TupleClass.__tupleArgs__
        if isinstance(p, PolymorphicTupleTypeFieldArg)
    ]
    if not polyType:
        return TupleClass

    fieldName = polyType[0].fieldName
    ConcreteTupleClass = TUPLE_TYPES_BY_NAME[jsonDict[fieldName]]

    if not issubclass(ConcreteTupleClass, TupleClass):
        raise TypeError(
            f"{ConcreteTupleClass.__name__} is not derived"
            f" from {TupleClass.__name__}"
        )

    return ConcreteTupleClass


class _TupleToPlainJsonMixin:
    def tupleToSmallJsonDict(self, includeNones=True, includeFalse=True):
        if not self.__shortFieldNamesMap__:
            raise Exception(
                "Tuple %s has no shortFieldNames defined" % self.tupleType()
            )

        return self.__tupleToJsonDict(
            includeNones=includeNones,
            includeFalse=includeFalse,
            useShortNames=True,
        )

    def tupleToRestfulJsonDict(self, includeNones=True, includeFalse=True):
        return self.__tupleToJsonDict(
            includeNones=includeNones,
            includeFalse=includeFalse,
            useShortNames=False,
        )

    def __tupleToJsonDict(
        self, includeNones=True, includeFalse=True, useShortNames=True
    ):

        if useShortNames:
            json = {
                "_tt": (
                    self.__tupleTypeShort__
                    if self.__tupleTypeShort__
                    else self.__tupleType__
                )
            }
        else:
            json = {}

        def convert(value):
            if isinstance(value, list):
                return [convert(v) for v in value]

            elif isinstance(value, dict):
                return {convert(k): convert(i) for k, i in value.items()}

            elif isinstance(value, Tuple):
                if useShortNames:
                    return value.tupleToSmallJsonDict()
                return value.tupleToRestfulJsonDict()

            elif isinstance(value, WKBElement):
                return convertFromWkbElement(value)

            elif isinstance(value, TupleField):
                return None

            elif isinstance(value, datetime):
                return value.strftime(ISO8601)

            else:
                return value

        for shortName, normalName in self.__shortFieldNamesMap__.items():
            value = convert(getattr(self, normalName))
            if value is None and not includeNones:
                continue
            if value is False and not includeFalse:
                continue
            if useShortNames:
                json[shortName] = value
            else:
                json[normalName] = value

        return json

    @staticmethod
    def smallJsonDictToTuple(jsonDict: dict) -> "Tuple":
        tupleShortType = jsonDict.get("_tt")
        if not tupleShortType:
            raise Exception(
                "Tuple.smallJsonDictToTuple: jsonDict has no _tt field"
            )

        Tuple_ = TUPLE_TYPES_BY_SHORT_NAME.get(tupleShortType)
        if not Tuple_:
            raise Exception(
                "Tuple.smallJsonDictToTuple: %s is not a registered tuple type"
                % tupleShortType
            )

        if not Tuple_.__shortFieldNamesMap__:
            raise Exception(
                "Tuple %s has no shortFieldNames defined" % Tuple.tupleType()
            )

        def convert(value):
            if value in (None, ""):
                return value

            if isinstance(value, list):
                return [convert(v) for v in value]

            if isinstance(value, dict):
                if value.get("_tt", None):
                    return Tuple.smallJsonDictToTuple(value)
                return {convert(k): convert(i) for k, i in value.items()}

            # elif isinstance(value, WKBElement):
            #     return convertFromWkbElement(value)

            if isinstance(value, str) and re.match(ISO8601_REGEXP, value):
                return strptime(value, ISO8601)

            return value

        newTuple = Tuple_()

        for shortName, normalName in Tuple_.__shortFieldNamesMap__.items():
            setattr(newTuple, normalName, convert(jsonDict.get(shortName)))

        return newTuple

    @staticmethod
    def restfulJsonDictToTupleWithValidation(
        jsonDict: dict, TupleClass: Type["Tuple"]
    ) -> "Tuple":
        """Restful JsonDict to Tuple With Validation

        This method uses the typings on the fields in a tuple to validate the
        types of data
        """

        if not get_type_hints(TupleClass):
            raise TypeError(f"{TupleClass} has no annotations")

        # Create concrete class for Tuple
        TupleClass = getConcreteClass(TupleClass, jsonDict)

        typeHints = get_type_hints(TupleClass)
        extraTypeHits = get_type_hints(TupleClass, include_extras=True)

        def getAndCheckTypeHint(fieldName):
            typeHint = typeHints.get(fieldName, None)

            if not typeHint:
                raise TypeError(
                    f"Type hint for {TupleClass} field {fieldName} is falsy"
                )

            # Check some type hints
            if get_origin(typeHint) is dict:
                keyHint, _ = get_args(typeHint)
                if keyHint is not str:
                    raise TypeError(
                        f"Type hint for {TupleClass}"
                        f" field {fieldName}"
                        f" must have a str type for the key,"
                        f" got {keyHint}"
                    )

            return typeHint

        def getExtraTypeHint(fieldName):
            extraTypeHit = extraTypeHits.get(fieldName, None)
            return extraTypeHit

        def getFieldValidatorFromExtraHint(extraTypeHint):

            origin = get_origin(extraTypeHint)

            # Check if the type hint is an Annotated
            if origin is not Annotated:
                return None

            for hint in get_args(extraTypeHint):
                if isinstance(hint, TupleFieldValidatorABC):
                    return hint

            return None

        def raiseIfNotType(
            fieldName,
            typeHint,
            value,
            validator: Optional[TupleFieldValidatorABC] = None,
        ):
            # if the field type is of the type typeHint, then we are good.
            if not get_origin(typeHint) and isinstance(value, typeHint):
                if validator:
                    validator.validate(fieldName, value)
                return

            # if the field type is of the type typeHint, then we are good.
            if isinstance(value, list) and get_origin(typeHint) is list:
                return

            # if the field type is of the type typeHint, then we are good.
            if isinstance(value, dict):
                if get_origin(typeHint) is dict:
                    return

                # If the type is a tuple and the value is a dict, that's a pass
                if hasattr(typeHint, "__tupleType__"):
                    return

            # handle the the Optional[fieldType] or Union case
            if type(value) in get_args(typeHint):
                return

            # Handle the None and Optional[fieldType] case
            if value is None and type(None) in get_args(typeHint):
                return

            raise TypeError(
                f"Parsing {TupleClass}, field {fieldName}"
                f" is not of type {typeHint}, got {type(value)} instead"
            )

        def convert(fieldName: str, value):
            typeHint = getAndCheckTypeHint(fieldName)
            extraTypeHint = getExtraTypeHint(fieldName)
            validator = getFieldValidatorFromExtraHint(extraTypeHint)

            raiseIfNotType(fieldName, typeHint, value, validator)

            if value in (None, ""):
                return value

            if isinstance(value, str) and re.match(ISO8601_REGEXP, value):
                return strptime(value, ISO8601)

            if isinstance(value, list):
                listItemTypeHint = get_args(typeHint)[0]
                listItemTypeValidator = getFieldValidatorFromExtraHint(
                    get_args(extraTypeHint)[0]
                )
                data = []
                if hasattr(listItemTypeHint, "__tupleType__"):
                    for item in value:
                        if not isinstance(item, dict):
                            raise TypeError(
                                f"Parsing {TupleClass}, field {fieldName}"
                                f" list item is not a json object to convert to "
                                f" {listItemTypeHint}, got {type(value)} instead"
                            )

                        data.append(
                            TupleClass.restfulJsonDictToTupleWithValidation(
                                item, listItemTypeHint
                            )
                        )

                else:
                    for item in value:
                        raiseIfNotType(
                            fieldName,
                            listItemTypeHint,
                            item,
                            listItemTypeValidator,
                        )
                        data.append(item)

                return data

            if isinstance(value, dict):
                # If this is just a single child tuple, import it
                if hasattr(typeHint, "__tupleType__"):
                    return TupleClass.restfulJsonDictToTupleWithValidation(
                        value, typeHint
                    )

                if not get_args(typeHint):
                    raise Exception(
                        "Typing should use `Dict[<type>,<type>]`," " not `dict`"
                    )

                keyTypeHint, valueTypeHint = get_args(typeHint)

                data = {}
                if hasattr(valueTypeHint, "__tupleType__"):
                    for key, item in value.items():
                        raiseIfNotType(fieldName, keyTypeHint, key)
                        if not isinstance(item, dict):
                            raise TypeError(
                                f"Parsing {TupleClass}, field {fieldName}"
                                f" list item is not a json object to convert to "
                                f" {valueTypeHint}, got {type(value)} instead"
                            )

                        data[
                            key
                        ] = TupleClass.restfulJsonDictToTupleWithValidation(
                            item, valueTypeHint
                        )

                else:
                    keyTypeExtraHint, valueTypeExtraHint = get_args(
                        extraTypeHint
                    )
                    keyTypeValidator = getFieldValidatorFromExtraHint(
                        keyTypeExtraHint
                    )
                    valueTypeValidator = getFieldValidatorFromExtraHint(
                        valueTypeExtraHint
                    )

                    for key, item in value.items():
                        raiseIfNotType(
                            fieldName, keyTypeHint, key, keyTypeValidator
                        )
                        raiseIfNotType(
                            fieldName, valueTypeHint, item, valueTypeValidator
                        )
                        data[key] = item  # No change

                return data

            # elif isinstance(value, WKBElement):
            #     return convertFromWkbElement(value)

            return value

        newTuple = TupleClass()

        for fieldName in TupleClass.__fieldNames__:
            setattr(
                newTuple, fieldName, convert(fieldName, jsonDict.get(fieldName))
            )

        return newTuple


class Tuple(Jsonable, _TupleToPlainJsonMixin, _TupleToSqlaJsonMixin):
    """Tuple

    This class provides rich serialisation / deserialisation support.

    The advantage over pure JSON is as follows:

    * Data structures are restored into full Tuple classes, allowing the use of
      properties, and methods on the tuples.

    * Datetimes and ints are serialised and deserialsied

    """

    #: Tuple Type, EG com.synerty.rapui.UnitTestTuple
    __tupleType__: str = None
    __tupleTypeShort__ = None
    __tupleArgs__ = ()  # Empty tuple
    __fieldNames__ = None
    __shortFieldNamesMap__ = None
    __rapuiSerialiseType__ = T_RAPUI_TUPLE

    def __init__(self, **kwargs):
        if self.__fieldNames__ is None:
            raise Exception("This tuple is missing the @addTupleType decorator")

        Jsonable.__init__(self)

        # If we're using slots, then don't use tuple fields (there arn't any anyway)
        if hasattr(self.__class__, "__slots__"):
            # noinspection PyTypeChecker
            for key in self.__class__.__slots__:
                if key in kwargs:
                    setattr(self, key, kwargs.pop(key))
                else:
                    setattr(self, key, None)

            if kwargs:
                raise KeyError(
                    "kwargs %s were passed, but tuple %s has no such fields"
                    % (", ".join(kwargs), self.__tupleType__)
                )

            return

        # Reset all the tuples.
        # We never want TupleField in an instance
        for name in self.__fieldNames__:
            tupleField = getattr(self.__class__, name)

            if isinstance(tupleField, TupleField):
                setattr(self, name, deepcopy(tupleField.defaultValue))

            elif isinstance(tupleField, InstrumentedAttribute):
                default = (
                    self.__table__.c[name].default
                    if name in self.__table__.c
                    else None
                )

                assign = (
                    default is not None
                    and getattr(self, name) is None
                    and not isinstance(default, Sequence)
                )
                if assign:
                    setattr(self, name, deepcopy(default.arg))

        # It's faster to add these at the end, rather than have logic to add one or
        # the other
        for key, val in kwargs.items():
            if not hasattr(self, key):
                raise KeyError(
                    "kwarg %s was passed, but tuple %s has no such TupleField"
                    % (key, self.__tupleType__)
                )
            setattr(self, key, val)

        # Handle setting any polymorphic field types
        for arg in self.__tupleArgs__:
            if isinstance(arg, PolymorphicTupleTypeFieldArg):
                setattr(self, arg.fieldName, self.__tupleType__)

    @classmethod
    def tupleFieldNames(cls) -> List[str]:
        return cls.__fieldNames__

    @classmethod
    def tupleName(cls):  # DEPRECIATED
        return cls.__tupleType__

    @classmethod
    def tupleType(cls):
        return cls.__tupleType__

    @classmethod
    def isSameTupleType(cls, other):
        if not hasattr(other, "__tupleType__"):
            return False
        return cls.__tupleType__ == other.__tupleType__

    def tupleClone(self):
        clone = self.__class__()
        for name in self.__fieldNames__:
            val = getattr(self, name)
            if val is not None:
                setattr(clone, name, val)

        return clone

    def _fromJson(self, jsonStr: str):
        jsonDict = json.loads(jsonStr)

        assert jsonDict[Jsonable.JSON_CLASS_TYPE] == self.__rapuiSerialiseType__
        return self.fromJsonDict(jsonDict)

    def _toJson(self) -> str:
        return json.dumps(self.toJsonDict())

    def __eq__(self, other):
        return id(self) == id(other)
        # if other == None or type(self) != type(other):
        #   return False
        #
        # # return id(self) != id(other)
        # for f in self.__fieldNames__:
        #   v1 = getattr(self, f, None)
        #   v2 = getattr(other, f, None)
        #
        #   if isinstance(v1, Tuple):
        #     if v1 is not v2:
        #       return False
        #   elif v1 != v2:
        #     return False
        #
        # return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return id(self)

    def __repr__(self):
        val = lambda name: (
            getattr(self, name) if hasattr(self, name) else "AttributeError"
        )

        vals = ["type = %s," % self.tupleType()]
        vals.extend(
            ["%s = %s," % (name, val(name)) for name in self.__fieldNames__]
        )

        return "\n".join(vals)


class TupleHash(object):
    def __init__(self, tupl):
        self.tupl = tupl
        from .Payload import Payload

        assert tupl is not None
        assert isinstance(tupl, Tuple) or isinstance(tupl, Payload)

    def _key(self):
        vals = []
        if isinstance(self.tupl, Tuple):
            vals += [self.tupl.tupleName()]
        for name in self.tupl.__fieldNames__:
            val = getattr(self.tupl, name)
            if isinstance(val, dict):
                newItems = []
                for k, v in list(val.items()):
                    k = TupleHash(k) if isinstance(k, Tuple) else k
                    v = TupleHash(v) if isinstance(v, Tuple) else v
                    newItems.append((k, v))
                val = tuple(newItems)

            elif isinstance(val, (list, set)):
                newItems = [
                    TupleHash(v)._key() if isinstance(v, Tuple) else v
                    for v in val
                ]
                val = tuple(newItems)

            elif isinstance(val, Tuple):
                val = TupleHash(val)._key()
                # val = TupleHash(val)

            vals.append((name, val))
        return tuple(vals)

    def __eq__(self, other):
        if not isinstance(other, TupleHash):
            return False
        return self._key() == other._key()

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        # This issue is caused by uiData and it's deep structure
        key = self._key()
        try:
            return hash(key)
        except Exception as e:
            raise e



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
import zlib
from base64 import b64encode, b64decode
from datetime import datetime
from typing import List, Optional, Dict

import pytz
import ujson

from vortex.DeferUtil import deferToThreadWrapWithLogger, noMainThread
from .Jsonable import Jsonable
from .SerialiseUtil import T_RAPUI_PAYLOAD

logger = logging.getLogger(__name__)

PayloadList = List["Payload"]
EncodedPayloadList = List[bytes]


class Payload(Jsonable):
    """Payload
    This object represents a hierarchy of data transferred between client and server
    """

    __fieldNames__ = ["filt", "tuples", "date"]
    __rapuiSerialiseType__ = T_RAPUI_PAYLOAD

    def __init__(self, filt: Optional[Dict] = None, tuples=None) -> None:
        """Constructor"""
        self.filt = {} if filt is None else filt
        self.tuples = [] if tuples is None else tuples
        self.date = datetime.now(pytz.utc)

        if isinstance(filt, str):
            self.filt = {self.filt: None}

        if not isinstance(self.tuples, list):
            self.tuples = [self.tuples]

    # -------------------------------------------
    # PayloadEnvelopes helper messages

    def makePayloadEnvelope(self, result=None, compressionLevel: int = 9):
        from .PayloadEnvelope import PayloadEnvelope

        noMainThread()
        encodedSelf = self.toEncodedPayload(compressionLevel=compressionLevel)
        return PayloadEnvelope(
            self.filt, encodedPayload=encodedSelf, date=self.date, result=result
        )

    @deferToThreadWrapWithLogger(logger)
    def makePayloadEnvelopeDefer(self, result=None, compressionLevel: int = 9):
        return self.makePayloadEnvelope(
            result=result, compressionLevel=compressionLevel
        )

    # -------------------------------------------
    # VortexMsg helper messages

    def makePayloadEnvelopeVortexMsg(
        self, result=None, compressionLevel: int = 9
    ):
        noMainThread()
        return self.makePayloadEnvelope(
            result=result, compressionLevel=compressionLevel
        ).toVortexMsg()

    @deferToThreadWrapWithLogger(logger)
    def makePayloadEnvelopeVortexMsgDefer(
        self, result=None, compressionLevel: int = 9
    ):
        return self.makePayloadEnvelopeVortexMsg(
            result=result, compressionLevel=compressionLevel
        )

    # -------------------------------------------
    # JSON Related methods
    def _fromJson(self, jsonStr):
        jsonDict = ujson.loads(jsonStr)

        assert jsonDict[Jsonable.JSON_CLASS_TYPE] == self.__rapuiSerialiseType__
        return self.fromJsonDict(jsonDict)

    def _toJson(self) -> str:
        return ujson.dumps(self.toJsonDict())

    # -------------------------------------------
    # VortexServer Message Methods
    def toEncodedPayload(self, compressionLevel: int = 9) -> bytes:
        jsonStr = self._toJson()
        return b64encode(
            zlib.compress(jsonStr.encode("UTF-8"), compressionLevel)
        )

    @deferToThreadWrapWithLogger(logger)
    def toEncodedPayloadDefer(self, compressionLevel: int = 9) -> bytes:
        return self.toEncodedPayload(compressionLevel=compressionLevel)

    def fromEncodedPayload(self, encodedPayload: bytes):
        jsonStr = zlib.decompress(b64decode(encodedPayload)).decode()
        return self._fromJson(jsonStr)

    @deferToThreadWrapWithLogger(logger)
    def fromEncodedPayloadDefer(self, encodedPayload: bytes):
        return self.fromEncodedPayload(encodedPayload)



import logging
from abc import abstractmethod, ABCMeta
from collections import namedtuple
from typing import Union, Optional, List

from twisted.internet import reactor
from twisted.internet.defer import fail, Deferred, succeed, inlineCallbacks
from twisted.python import failure

from vortex.DeferUtil import vortexLogFailure
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.TupleSelector import TupleSelector
from vortex.VortexABC import SendVortexMsgResponseCallable
from vortex.VortexFactory import VortexFactory

logger = logging.getLogger(__name__)


class TupleSelectorUpdateMapperABC(metaclass=ABCMeta):
    @abstractmethod
    def mapTupleSelector(
        self,
        triggerTupleSelector: TupleSelector,
        allTupleSelectors: List[TupleSelector],
    ) -> List[TupleSelector]:
        """Map Tuple Selector

        Implementing this class provides some intelligence to the Observable
        to trigger a refresh of related data to the triggerTupleSelector.

        :param triggerTupleSelector: The TupleSelector that has triggered the update.
        :param allTupleSelectors: All TupleSelectors being observed from this observable.
        """


class TuplesProviderABC(metaclass=ABCMeta):
    @abstractmethod
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        """Make Vortex Msg

        The method generates the vortexMsg for the vortex to send.

        :param filt: The filt for the payload.
        :param tupleSelector: The tuple selector us used to determing which tuples to
        send back,
        """


_ObserverDetails = namedtuple(
    "_ObserverDetails", ["vortexUuid", "observerName"]
)


class _ObserverData:
    def __init__(self, tupleSelector: TupleSelector):
        self.tupleSelector = tupleSelector
        self.observers = set()


class TupleDataObservableHandler:
    def __init__(
        self,
        observableName,
        additionalFilt=None,
        subscriptionsEnabled=True,
        acceptOnlyFromVortex: Optional[Union[str, tuple]] = None,
    ):
        """Constructor

        :param observableName: The name of this observable

        :param additionalFilt: Any additional filter keys that are required

        :param subscriptionsEnabled: Should subscriptions be enabled (default)

        :param acceptOnlyFromVortex: Accept requests only from this vortex,
            The vortex can be str or tuple of str, or None to accept from any.

        """
        self._observableName = observableName
        self._subscriptionsEnabled = subscriptionsEnabled
        self._filt = dict(name=observableName, key="tupleDataObservable")
        if additionalFilt:
            self._filt.update(additionalFilt)

        self._endpoint = PayloadEndpoint(
            self._filt, self._process, acceptOnlyFromVortex=acceptOnlyFromVortex
        )

        self._observerDataByTupleSelector = {}

        self._tupleProvidersByTupleName = {}
        self._tupleSelectorUpdateMappers = []

    def addTupleProvider(self, tupleName, provider: TuplesProviderABC):
        """Add Tuple Provider"""
        assert (
            not tupleName in self._tupleProvidersByTupleName
        ), "Observable:%s, Tuple name %s is already registered" % (
            self._observableName,
            tupleName,
        )

        assert isinstance(provider, TuplesProviderABC), (
            "Observable:%s, provider must be an instance of TuplesProviderABC"
            % self._observableName
        )

        self._tupleProvidersByTupleName[tupleName] = provider

    def addTupleSelectorUpdateMapper(
        self, mapper: TupleSelectorUpdateMapperABC
    ):
        """Add Tuple Selector Update Mapper

        This mapper will be called every time a tuple selector is notified of an update.

        """
        self._tupleSelectorUpdateMappers.append(mapper)

    def hasTupleProvider(self, tupleName: str):
        return tupleName in self._tupleProvidersByTupleName

    def hasTupleSubscribers(self, tupleSelector: TupleSelector):
        return tupleSelector.toJsonStr() in self._observerDataByTupleSelector

    def shutdown(self):
        self._endpoint.shutdown()
        self._tupleSelectorUpdateMappers = []
        self._tupleProvidersByTupleName = {}
        self._observerDataByTupleSelector = {}

    @inlineCallbacks
    def _createVortexMsg(self, filt, tupleSelector: TupleSelector) -> Deferred:
        tupleProvider = self._tupleProvidersByTupleName.get(tupleSelector.name)
        assert (
            tupleProvider
        ), "Observable:%s, No providers registered for tupleName %s" % (
            self._observableName,
            tupleSelector.name,
        )
        try:
            vortexMsg = yield self._customMaybeDeferred(
                tupleProvider.makeVortexMsg, filt, tupleSelector
            )
        except Exception as e:
            logger.exception(e)
            raise
        if not vortexMsg:
            msg = f"TupleDataProvider did not return VortexPY message {tupleSelector}"
            logger.error(msg)
            raise Exception(msg)
        return vortexMsg

    @inlineCallbacks
    def _process(
        self,
        payloadEnvelope: PayloadEnvelope,
        vortexUuid: str,
        sendResponse: SendVortexMsgResponseCallable,
        **kwargs,
    ):
        tupleSelector = payloadEnvelope.filt["tupleSelector"]
        tsStr = tupleSelector.toJsonStr()

        observerDetails = _ObserverDetails(
            vortexUuid, payloadEnvelope.filt.get("observerName")
        )

        # Handle unsubscribe
        if payloadEnvelope.filt.get("unsubscribe"):
            if tsStr in self._observerDataByTupleSelector:
                try:
                    self._observerDataByTupleSelector[tsStr].observers.remove(
                        observerDetails
                    )

                except KeyError:
                    pass

                if not self._observerDataByTupleSelector[tsStr].observers:
                    del self._observerDataByTupleSelector[tsStr]

            return

        # Add support for just getting data, no subscription.
        if self._subscriptionsEnabled and payloadEnvelope.filt.get(
            "subscribe", True
        ):
            if tsStr not in self._observerDataByTupleSelector:
                self._observerDataByTupleSelector[tsStr] = _ObserverData(
                    tupleSelector
                )

            self._observerDataByTupleSelector[tsStr].observers.add(
                observerDetails
            )

        vortexMsg = yield self._createVortexMsg(
            payloadEnvelope.filt, tupleSelector
        )
        try:
            yield sendResponse(vortexMsg)
        except Exception as e:
            logger.exception(e)

    def _getMappedTupleSelectors(
        self, tupleSelector: TupleSelector
    ) -> List[TupleSelector]:

        # Get all tuple selectors
        allTupleSelectors = [
            data.tupleSelector
            for data in self._observerDataByTupleSelector.values()
        ]

        # Create a dict so we end up with only unique ones
        tupleSelectorByStr = {tupleSelector.toJsonStr(): tupleSelector}

        # Run through the mappers
        for mapper in self._tupleSelectorUpdateMappers:
            mappedSelectors = mapper.mapTupleSelector(
                tupleSelector, allTupleSelectors
            )
            if mappedSelectors:
                tupleSelectorByStr.update(
                    {ts.toJsonStr(): ts for ts in mappedSelectors}
                )

        # Return a list of tuple selectors
        return list(tupleSelectorByStr.values())

    def notifyOfTupleUpdate(self, tupleSelector: TupleSelector) -> None:
        """Notify Of Tuple Update

        This method tells the observable that an update has occurred and that it should
        send new data to it's observers.

        Tuple selectors should be identical to the data being observed.

        :param tupleSelector: A tuple selector that describes the scope of the update.
        :returns: None

        """
        tupleSelectors = self._getMappedTupleSelectors(tupleSelector)
        reactor.callLater(0, self._notifyOfTupleUpdateInMain, tupleSelectors)

    def notifyOfTupleUpdateForTuple(self, tupleName: str) -> None:
        """Notify of Tuple Update for Tuple

        Like the above notification method, this method will aquire new data and
        send it to observers.

        .. note:: Calling this will not trigger the TupleSelector Update Mappers as
                  this is an earlier soloution.

                  For all new code, use notifyOfTupleUpdate() and define mappers.

        """
        tupleSelectors = []
        for data in self._observerDataByTupleSelector.values():
            if data.tupleSelector.name == tupleName:
                tupleSelectors.append(data.tupleSelector)

        reactor.callLater(0, self._notifyOfTupleUpdateInMain, tupleSelectors)

    @inlineCallbacks
    def _notifyOfTupleUpdateInMain(self, tupleSelectors: List[TupleSelector]):
        for tupleSelector in tupleSelectors:
            yield self._notifyOfTupleUpdateInMainOne(tupleSelector)

    @inlineCallbacks
    def _notifyOfTupleUpdateInMainOne(self, tupleSelector: TupleSelector):
        tsStr = tupleSelector.toJsonStr()
        if tsStr not in self._observerDataByTupleSelector:
            return

        # Filter out the offline observables
        onlineUuids = set(VortexFactory.getRemoteVortexUuids())
        observers = self._observerDataByTupleSelector[tsStr].observers
        for od in list(observers):
            if od.vortexUuid not in onlineUuids:
                observers.remove(od)

        # Get / update the list of observing UUIDs
        if not observers:
            del self._observerDataByTupleSelector[tsStr]
            return

        # Create the vortexMsg
        filt = dict(tupleSelector=tupleSelector)
        filt.update(self._filt)
        vortexMsg = yield self._createVortexMsg(filt, tupleSelector)

        # We can have multiple Observable clients on the one vortex, so make sure
        # we only send one message for these.
        destVortexUuids = set([od.vortexUuid for od in observers])

        # Send the vortex messages
        for destVortexUuid in destVortexUuids:
            d = VortexFactory.sendVortexMsg(
                vortexMsgs=vortexMsg, destVortexUuid=destVortexUuid
            )
            d.addErrback(vortexLogFailure, logger)

    def _customMaybeDeferred(self, f, *args, **kw):
        try:
            result = f(*args, **kw)
        except Exception as e:
            logger.error(
                "TupleDataObservableHandler:%s TupleProvider failed",
                self._observableName,
            )
            logger.exception(e)
            return fail(failure.Failure(e))

        if isinstance(result, Deferred):
            return result
        elif isinstance(result, failure.Failure):
            return fail(result)
        else:
            return succeed(result)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : https://www.synerty.com
 * Support : support@synerty.com
"""
import logging
from typing import Type, Dict, Union

from twisted.internet.defer import Deferred, inlineCallbacks

from vortex.DeferUtil import callMethodLater
from vortex.DataLoaderDelegate import TupleDataLoaderDelegateABC
from vortex.Payload import Payload
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.Tuple import TupleField, addTupleType, Tuple
from vortex.TupleAction import TupleActionABC
from vortex.TupleActionVortex import TupleActionVortex
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleActionProcessor import (
    TupleActionProcessorDelegateABC,
    TupleActionProcessor,
)
from vortex.handler.TupleDataObservableHandler import (
    TupleDataObservableHandler,
    TuplesProviderABC,
)

logger = logging.getLogger(__name__)


class _DataLoaderDelegateState:
    def __init__(self, delegate):
        self.delegate = delegate
        # TODO: Make this a vortexMsgByTs instead of liveValueByTs
        self.liveValueByTs = {}


STORE_ACTION = "store"
EDIT_UPDATE_ACTION = "edit_update"
LOAD_ACTION = "load"
DELETE_ACTION = "delete"


@addTupleType
class _DataLoaderTupleAction(TupleActionABC):
    __tupleType__ = "vortex._DataLoaderTupleAction"

    data: Type[Tuple] = TupleField()
    action: str = TupleField()
    selector: TupleSelector = TupleField()


@addTupleType
class _LockDataTupleAction(TupleActionABC):
    __tupleType__ = "vortex._LockDataTupleAction"

    selector: TupleSelector = TupleField()
    lock: bool = TupleField()


@addTupleType
class _DataLockStatusTuple(Tuple):
    __tupleType__ = "vortex._DataLockStatusTuple"

    locked: bool = TupleField()


class TupleDataLoader(TupleActionProcessorDelegateABC, TuplesProviderABC):
    """Tuple Data Loader

    The `TupleDataLoader class is responsible for maintaining the
    `TupleDataObservableHandler` and `TupleActionProcessor` pair for each module
    which uses TupleForms. Forms (`Tuple`) and their corresponding
    delegate (`TupleDataLoaderDelegateABC`) are registered with the modules form loader
    using the `setDelegate` method. `TupleDataLoader`s lifecycle should be
    managed with the `shutdown` method in the UI controllers
    `shutdown` method. The `TupleDataLoader` can (should??) be initialized in
    the `start` method of the UI controller.

    Each instance of the `TupleDataLoader` registers itself as the delegate for
    every form (which itself is a `TupleActionABC`) registered with it. The
    `TupleDataLoaderDelegateABC` methods are used for handling reads and writes.

    TODO: Locking

    """

    def __init__(
        self,
        observable: TupleDataObservableHandler,
        actionProcessor: TupleActionProcessor,
    ):
        self._observable = observable
        self._actionProcessor = actionProcessor

        self._registeredDataDelegateByDataType: Dict[
            str, _DataLoaderDelegateState
        ] = {}

        self._lockingUuidByTs: Dict[TupleSelector, str] = {}

    def start(self):
        """Start

        Start all registered handlers

        """
        self._observable.addTupleProvider(
            _DataLockStatusTuple.tupleType(), self
        )

        self._actionProcessor.setDelegate(
            _DataLoaderTupleAction.tupleType(),
            self,
        )
        self._actionProcessor.setDelegate(
            _LockDataTupleAction.tupleType(), self
        )

        for entry in self._registeredDataDelegateByDataType.values():
            entry.delegate.start()

    def shutdown(self) -> None:
        """Shutdown

        Shutdown the observable and action processor managed with this
        TupleDataLoader and the registered handlers

        :return: None
        """
        for entry in self._registeredDataDelegateByDataType.values():
            entry.delegate.shutdown()

        self._observable.shutdown()
        self._actionProcessor.shutdown()

    def setDelegate(
        self,
        DataClass: Union[Type[Tuple], str],
        delegate: TupleDataLoaderDelegateABC,
    ) -> None:
        """Set Delegate

        Sets `self` as the delegate for `DataClass` which is a TupleActionABC
        and registers `handler` as the form data handler for `FormClass` which
        is used for reading and writing the data

        :param DataClass: `Tuple` to register the handler for
        :param delegate: Instance of `TupleDataLoaderDelegateABC` that provides R/W
        :return: None
        """
        tupleType = (
            DataClass if isinstance(DataClass, str) else DataClass.tupleType()
        )

        self._observable.addTupleProvider(tupleType, self)
        self._registeredDataDelegateByDataType[
            tupleType
        ] = _DataLoaderDelegateState(delegate=delegate)

    @inlineCallbacks
    def makeVortexMsg(
        self, filt: dict, tupleSelector: TupleSelector
    ) -> Union[Deferred, bytes]:
        if tupleSelector.name == _DataLockStatusTuple.tupleType():
            selector = TupleSelector().fromJsonDict(tupleSelector.selector)

            result = _DataLockStatusTuple()
            result.locked = False

            if selector in self._lockingUuidByTs:
                result.locked = True
        else:
            tupleType = tupleSelector.name

            if tupleType not in self._registeredDataDelegateByDataType:
                raise ValueError(
                    f"Delegate for {tupleType} not registered with "
                    f"{self.__class__}"
                )

            entry = self._registeredDataDelegateByDataType[tupleType]
            if entry.liveValueByTs.get(tupleSelector) is None:
                result = yield entry.delegate.loadData(tupleSelector)
            else:
                result = entry.liveValueByTs.get(tupleSelector)

        results = [result]

        payloadEnvelope = PayloadEnvelope(filt=filt)
        payloadEnvelope.encodedPayload = Payload(
            filt=filt, tuples=results
        ).toEncodedPayload()
        vortexMsg = payloadEnvelope.toVortexMsg()

        return vortexMsg

    @inlineCallbacks
    def processTupleAction(
        self,
        tupleAction: _DataLoaderTupleAction,
        tupleActionVortex: TupleActionVortex,
    ) -> Deferred:
        lockingTs = TupleSelector(
            name=_DataLockStatusTuple.tupleType(),
            selector={
                "name": tupleAction.selector.name,
                "selector": tupleAction.selector.selector,
            },
        )

        if isinstance(tupleAction, _LockDataTupleAction):
            if tupleAction.lock:
                if tupleAction.selector in self._lockingUuidByTs:
                    raise Exception("Data already locked")

                self._lockingUuidByTs[
                    tupleAction.selector
                ] = tupleActionVortex.uuid
            else:
                if tupleAction.selector not in self._lockingUuidByTs:
                    raise Exception("Data not locked")

                lockingUuid = self._lockingUuidByTs[tupleAction.selector]

                if lockingUuid != tupleActionVortex.uuid:
                    raise Exception("Only the locker can unlock the data")

                self._lockingUuidByTs.pop(tupleAction.selector)
            self.__notifyOfTupleUpdate(lockingTs)
            return

        if (
            tupleAction.data.tupleType()
            not in self._registeredDataDelegateByDataType
        ):
            raise ValueError(
                f"Delegate for {tupleAction.data.tupleType()} not registered with "
                f"{self.__class__}"
            )

        entry = self._registeredDataDelegateByDataType[
            tupleAction.data.tupleType()
        ]

        if tupleAction.selector in self._lockingUuidByTs:
            if (
                tupleActionVortex.uuid
                != self._lockingUuidByTs[tupleAction.selector]
            ):
                raise Exception("Data is locked")

        if tupleAction.action == STORE_ACTION:
            yield entry.delegate.storeData(
                tupleAction.data, tupleAction.selector.selector
            )
            entry.liveValueByTs[tupleAction.selector] = None

            # Unlock the form if it is locked
            if tupleAction.selector in self._lockingUuidByTs:
                self._lockingUuidByTs.pop(tupleAction.selector)
                self.__notifyOfTupleUpdate(lockingTs)

        elif tupleAction.action == LOAD_ACTION:
            # It looks like nothing is done for LOAD_ACTION but the work is done
            # in the makeVortexMsg method
            entry.liveValueByTs[tupleAction.selector] = None

            # Unlock the form if it is locked
            if tupleAction.selector in self._lockingUuidByTs:
                self._lockingUuidByTs.pop(tupleAction.selector)
                self.__notifyOfTupleUpdate(lockingTs)

        elif tupleAction.action == DELETE_ACTION:
            yield entry.delegate.deleteData(tupleAction.selector.selector)
            entry.liveValueByTs[tupleAction.selector] = None

            # Unlock the form if it is locked
            if tupleAction.selector in self._lockingUuidByTs:
                self._lockingUuidByTs.pop(tupleAction.selector)
                self.__notifyOfTupleUpdate(lockingTs)

        elif tupleAction.action == EDIT_UPDATE_ACTION:
            entry.liveValueByTs[tupleAction.selector] = tupleAction.data

        self.__notifyOfTupleUpdate(tupleAction.selector)

    @callMethodLater
    def __notifyOfTupleUpdate(self, ts: TupleSelector):
        self._observable.notifyOfTupleUpdate(ts)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import base64
import logging
from typing import Any
from typing import Union

import ujson
import zlib
from base64 import b64encode, b64decode
from datetime import datetime
from typing import List, Optional, Dict

import pytz

from .DeferUtil import deferToThreadWrapWithLogger, noMainThread
from .Payload import Payload
from .SerialiseUtil import T_RAPUI_PAYLOAD_ENVELOPE
from .Jsonable import Jsonable

logger = logging.getLogger(__name__)

PayloadEnvelopeList = List["PayloadEnvelope"]
VortexMsgList = List[bytes]


class NoPayloadException(Exception):
    pass


class PayloadEnvelope(Jsonable):
    """Payload
    This object represents a hierarchy of data transferred between client and server
    """

    __fieldNames__ = ["filt", "data", "result", "date"]
    __rapuiSerialiseType__ = T_RAPUI_PAYLOAD_ENVELOPE

    vortexUuidKey = "__vortexUuid__"
    vortexNameKey = "__vortexName__"

    def __init__(
            self,
            filt: Optional[Dict] = None,
            encodedPayload: Optional[str] = None,
            result=None,
            date: Optional[datetime] = None,
            data: Optional[Any] = None,
    ) -> None:
        Jsonable.__init__(self)
        """Constructor"""
        assert not (
                data and encodedPayload
        ), "You can not pass encodedData and encodedPayload"

        self.filt: Dict = {} if filt is None else filt
        self.data: Union[Any, str, None] = data if data else encodedPayload
        self.result = result
        self.date: datetime = date if date else datetime.now(pytz.utc)

        if isinstance(filt, str):
            self.filt = {self.filt: None}

    # -------------------------------------------
    # PayloadEnvelope Methods

    @property
    def encodedPayload(self) -> Optional[str]:
        # self.data could be from VortexJS/Typescript where we allow `str` data
        assert (
                self.data is None
                or isinstance(self.data, str)
                or isinstance(self.data, bytes)
        ), "Encoded payload is not None, bytes or string"
        return self.data

    @encodedPayload.setter
    def encodedPayload(self, val: Optional[str]):
        # This value should only be set from users of VortexPy and
        # Payload.toEncodedPayload so we only allow bytes or None
        assert val is None or isinstance(
            val, bytes
        ), "Encoded payload is not None or a string"
        self.data = val

    def isEmpty(self):
        return (
                (
                        not self.filt
                        or (
                                self.vortexNameKey in self.filt
                                and self.vortexUuidKey in self.filt
                                and len(self.filt) == 2
                        )
                )
                and not self.encodedPayload
                and not self.result
        )

    # -------------------------------------------
    # Decode Helper Methods

    def decodePayload(self) -> Payload:
        if not self.encodedPayload:
            if self.result:
                logger.debug(
                    "encodedPayload is None, but maybe there was an error: %s",
                    self.result,
                )
            raise NoPayloadException()

        noMainThread()
        return Payload().fromEncodedPayload(self.encodedPayload)

    @deferToThreadWrapWithLogger(logger)
    def decodePayloadDefer(self) -> "Payload":
        return self.decodePayload()

    # -------------------------------------------
    # JSON Related methods
    def _fromJson(self, jsonStr):
        jsonDict = ujson.loads(jsonStr)

        assert jsonDict[Jsonable.JSON_CLASS_TYPE] == self.__rapuiSerialiseType__
        return self.fromJsonDict(jsonDict)

    def _toJson(self) -> str:
        return ujson.dumps(self.toJsonDict())

    # -------------------------------------------
    # VortexServer Message Methods
    def toVortexMsg(self, base64Encode=False) -> bytes:
        jsonStr = self._toJson().encode()
        if base64Encode:
            return b64encode(jsonStr)
        return jsonStr

    @deferToThreadWrapWithLogger(logger)
    def toVortexMsgDefer(self, base64Encode=False) -> bytes:
        return self.toVortexMsg(base64Encode=base64Encode)

    def fromVortexMsg(
            self,
            vortexMsg: bytes,
    ) -> "PayloadEnvelope":
        if vortexMsg.startswith(b"{"):
            jsonStr = vortexMsg.decode()
        else:
            jsonStr = b64decode(vortexMsg).decode()
        return self._fromJson(jsonStr)

    @deferToThreadWrapWithLogger(logger)
    def fromVortexMsgDefer(self, vortexMsg: bytes) -> "PayloadEnvelope":
        return self.fromVortexMsg(vortexMsg)

    @classmethod
    @deferToThreadWrapWithLogger(logger)
    def base64EncodeDefer(cls, vortexMsg) -> bytes:
        return base64.b64encode(vortexMsg)



from vortex.rpc.RPC import vortexRPC


@vortexRPC("listenVortexName", acceptOnlyFromVortex="sendVortexName")
def myRemoteAddMethod(arg1: int, kwarg1: int = 0) -> int:
    print(arg1, kwarg1)
    raise ValueError("A is not B")



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : https://www.synerty.com
 * Support : support@synerty.com
"""
import abc
from abc import abstractmethod
from typing import Union, Dict, Any

from twisted.internet.defer import Deferred, inlineCallbacks

from vortex.Tuple import Tuple
from vortex.TupleSelector import TupleSelector


class TupleDataLoaderDelegateABC(abc.ABC):
    """Tuple Data Loader Delegate ABC

    ABC for delegate classes to handle the reading and writing for each forms'
    data. Handlers must be register in
    (`Tuple`, `TupleDataLoaderDelegateABC`) pairs with the `TupleDataLoader`

    """

    def start(self):
        pass

    def shutdown(self):
        pass

    @abstractmethod
    def loadData(self, selector: TupleSelector) -> Union[Deferred, Tuple]:
        """Load Data

        The `loadData` method is called when data needs to be loaded from a
        storage medium. The query is represented by a `TupleSelector` value
        `selector`.

        :param selector: The `TupleSelector` for which to load data
        :return: The `Tuple` or a `Deferred` which eventually results in a `Tuple`
        """
        pass

    @abstractmethod
    def storeData(self, data: Tuple, selector: TupleSelector) -> Deferred:
        """Store Data

        Store the data into the storage medium. The `data` is the `Tuple` which
        needs to be written

        :param data: Tuple to write and store
        :param selector: Previous selector used
        :return: Deferred
        """
        pass

    @abstractmethod
    def deleteData(self, selector: TupleSelector) -> Deferred:
        """Delete Data

        Delete the data referenced by the `selector`

        :param selector: `TupleSelector` for data to be deleted
        :return: Deferred
        """
        pass



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
from abc import ABCMeta, abstractmethod
from collections import deque

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from twisted.internet.error import ConnectionDone, ConnectionLost
from twisted.internet.protocol import Protocol, connectionDone

from vortex.DeferUtil import nonConcurrentMethod
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.PayloadIO import PayloadIO

logger = logging.getLogger(name=__name__)


class VortexPayloadProtocol(Protocol, metaclass=ABCMeta):
    def __init__(self, logger):
        self._data = b""
        self._serverVortexUuid = None
        self._serverVortexName = None
        self._logger = logger

        self._processVortexMsgsInProgress = False
        self._vortexMsgsQueue = deque()

    @abstractmethod
    def _beat(self):
        """
        EG :
        if self._vortexClient:
            self._vortexClient._beat()
        """

    @abstractmethod
    def _nameAndUuidReceived(self, name, uuid):
        """
        EG :
        if self._vortexClient:
            self._vortexClient._setNameAndUuid(name=self._serverVortexName,
                                               uuid=self._serverVortexUuid)
        """

    @abstractmethod
    def _createResponseSenderCallable(self):
        """
        EG :
        def sendResponse(vortexMsgs: Union[VortexMsgList, bytes]):
            return self._vortexClient.sendVortexMsg(vortexMsgs=vortexMsgs)
        """

    def vortexMsgReceived(self, vortexMsg):
        self._data += vortexMsg
        self._beat()
        if vortexMsg == b".":
            return

        self._vortexMsgsQueue.append(vortexMsg)

        if self._vortexMsgsQueue and not self._processVortexMsgs.running:
            reactor.callLater(0, self._processVortexMsgs)

    def dataReceived(self, bytesIn):
        if bytesIn.startswith(b"<"):
            raise Exception("Not Logged In")

        self._data += bytesIn
        self._beat()
        self._processData()

    def connectionLost(self, reason=connectionDone):
        reasonFailure = reason
        self._processData()
        if reasonFailure.check(ConnectionDone):
            self._logger.info(
                "Connection closed by other end (it may be shutting down)"
            )

        elif isinstance(reasonFailure.value, ConnectionLost):
            self._logger.info(
                "Connection to other end lost (We may be shutting down)"
            )

        else:
            self._logger.error("Closed with error")
            try:
                self._logger.exception(reason.getErrorMessage())
                self._logger.exception(reason.getTraceback())
            except:
                self._logger.exception(reasonFailure.value)

    def _processData(self):
        if not self._data:
            return

        def getNextChunkIter():
            try:
                while True:
                    yield self._data.index(b".")
            except ValueError:
                # There is no '.' in it, wait for more data.
                return

        for nextChunk in getNextChunkIter():
            vortexMsg = self._data[:nextChunk]
            self._data = self._data[nextChunk + 1:]

            # If we get two heartbeats in a row, this will be false
            if len(vortexMsg):
                self._vortexMsgsQueue.append(vortexMsg)

        if self._vortexMsgsQueue and not self._processVortexMsgs.running:
            reactor.callLater(0, self._processVortexMsgs)

    @inlineCallbacks
    @nonConcurrentMethod
    def _processVortexMsgs(self):
        while self._vortexMsgsQueue:
            vortexMsg = self._vortexMsgsQueue.popleft()

            try:
                payloadEnvelope = yield PayloadEnvelope().fromVortexMsgDefer(
                    vortexMsg
                )

                if payloadEnvelope.isEmpty():
                    self._processServerInfoPayload(payloadEnvelope)
                else:
                    self._deliverPayload(payloadEnvelope)

            except Exception as e:
                print(vortexMsg)
                print(e)
                self._logger.exception(e)
                raise

    def _processServerInfoPayload(self, payload):
        """Process Server Info Payload

        The first payload a server sends to the client contains information about it's
        self.

        """

        if PayloadEnvelope.vortexUuidKey in payload.filt:
            self._serverVortexUuid = payload.filt[PayloadEnvelope.vortexUuidKey]

        if PayloadEnvelope.vortexNameKey in payload.filt:
            self._serverVortexName = payload.filt[PayloadEnvelope.vortexNameKey]

        self._nameAndUuidReceived(
            name=self._serverVortexName, uuid=self._serverVortexUuid
        )

    def _deliverPayload(self, payload):

        PayloadIO().process(
            payload,
            vortexUuid=self._serverVortexUuid,
            vortexName=self._serverVortexName,
            httpSession=None,
            sendResponse=self._createResponseSenderCallable(),
        )



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
from urllib.parse import urlparse, parse_qs

from twisted.internet.defer import inlineCallbacks
from twisted.internet.protocol import Protocol, connectionDone, Factory
from twisted.protocols.policies import WrappingFactory
from twisted.web import resource
from twisted.web.server import NOT_DONE_YET
from txwebsocket.txws import WebSocketProtocol

from vortex.DeferUtil import vortexLogFailure
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexServerConnection import VortexServerConnection
from .PayloadPriority import DEFAULT_PRIORITY
from .VortexServer import VortexServer

logger = logging.getLogger(name=__name__)


class VortexWebsocketServerProtocol(Protocol):
    def __init__(self, vortex: VortexServer, addr, httpSession=None) -> None:
        self._vortex = vortex
        self._addr = addr

        self._conn = None

        self._remoteVortexUuid = None
        self._remoteVortexName = None
        self._httpSession = httpSession

        def httpSessionExpiredCallback():
            self._httpSession = None
            self.transport.loseConnection()

        if self._httpSession:
            self._httpSession.notifyOnExpire(httpSessionExpiredCallback)

        # Most messages don't need a buffer, but websockets can split messages
        # around the 128kb mark
        self._receiveBuffer = b""

    def __initConnection(self):
        # self.transport.setBinaryMode(True)

        params = parse_qs(urlparse(self.transport.location).query)

        if "vortexUuid" not in params or "vortexName" not in params:
            raise Exception(
                "This isn't a vortex capable websocket. Check the URL"
            )

        self._remoteVortexUuid = params["vortexUuid"][0]
        self._remoteVortexName = params["vortexName"][0]
        self._conn = VortexServerConnection(
            self._vortex,
            self._remoteVortexUuid,
            self._remoteVortexName,
            self._httpSession,
            self.transport,
            self._addr,
        )

        # Send a heart beat down the new connection, tell it who we are.
        connectPayloadFilt = {}
        connectPayloadFilt[PayloadEnvelope.vortexUuidKey] = self._vortex.uuid()
        connectPayloadFilt[PayloadEnvelope.vortexNameKey] = self._vortex.name()
        self._conn.write(
            PayloadEnvelope(filt=connectPayloadFilt).toVortexMsg(),
            DEFAULT_PRIORITY,
        )

        self._vortex.connectionOpened(self._httpSession, self._conn)

    def dataReceived(self, data):

        if not self._conn:
            self.__initConnection()

        if self._vortex.isShutdown():
            return None

        if self._httpSession:
            self._httpSession.touch()

        if data == b".":
            self._conn.beatReceived()
            return

        if len(data) > 1 and data[-1] == b".":
            data = data[:-1]

        d = self._processVortexMsg(data)
        d.addErrback(vortexLogFailure, logger, consumeError=True)

    def connectionLost(self, reason=connectionDone):
        if self._conn:
            self._conn.transportClosed()

    @inlineCallbacks
    def _processVortexMsg(self, vortexMsg: bytes):
        payloadEnvelope = yield PayloadEnvelope().fromVortexMsgDefer(vortexMsg)
        self._vortex.payloadReveived(
            httpSession=self._httpSession,
            vortexUuid=self._remoteVortexUuid,
            vortexName=self._remoteVortexName,
            payload=payloadEnvelope,
        )


class VortexWebsocketServerFactory(Factory):
    protocol = None

    def __init__(self, vortexServer: VortexServer) -> None:
        self._vortexServer = vortexServer

    def buildProtocol(self, addr, httpSession=None):
        p = VortexWebsocketServerProtocol(self._vortexServer, addr, httpSession)
        p.factory = self
        return p


class VortexWrappedWebSocketFactory(WrappingFactory):
    """
    Factory which wraps another factory to provide WebSockets transports for
    all of its protocols.
    """

    protocol = WebSocketProtocol

    def buildProtocol(self, addr, httpSession):
        return self.protocol(
            self, self.wrappedFactory.buildProtocol(addr, httpSession)
        )


class VortexWebSocketUpgradeResource(resource.Resource):
    """Vortex Websocket Upgrade Resource

    If this resource is hit, it will attempt to upgrade the connection to a websocket.

    """

    isLeaf = 1

    def __init__(self, websocketFactory):
        """Constructor

        @:param websocketFactory: A factory that will build a WebsocketProtocol (above)
        """
        resource.Resource.__init__(self)
        self._websocketFactory = websocketFactory

    def render(self, request):
        httpSession = request.getSession()

        websocketProtocol = self._websocketFactory.buildProtocol(
            request.client.host, httpSession
        )
        websocketProtocol.makeConnection(request.channel.transport)
        websocketProtocol.initFromRequest(request)
        request.channel.upgradeToWebsocket(websocketProtocol)

        return NOT_DONE_YET



import logging
from copy import copy

from twisted.internet.defer import Deferred

from vortex.Payload import Payload
from vortex.PayloadResponse import PayloadResponse
from vortex.TupleAction import TupleActionABC

logger = logging.getLogger(__name__)


class TupleDataActionClient:
    def __init__(
        self,
        destVortexName: str,
        tupleActionProcessorName: str,
        additionalFilt: dict = None,
    ) -> None:
        """Constructor

        :param destVortexName: The name of the destination vortex to send to.

        :param tupleActionProcessorName: The name of this observable

        :param additionalFilt: Any additional filter keys that are required
        """
        self._destVortexName = destVortexName
        self._filt = dict(
            name=tupleActionProcessorName, key="tupleActionProcessorName"
        )
        if additionalFilt:
            self._filt.update(additionalFilt)

    def pushAction(
        self, tupleAction: TupleActionABC, additionalFilt=None
    ) -> Deferred:
        """Push Action

        This pushes the action, either locally or to the server, depending on the
        implementation.

        If pushed locally, the promise will resolve when the action has been saved.
        If pushed directly to the server, the promise will resolve when the server has
        responded.

        :param tupleAction The tuple action to send to the remote end
        :param additionalFilt: Any extra filt that may be required to send with the
                payload.
        """
        filt = copy(self._filt)
        if additionalFilt:
            filt.update(additionalFilt)

        d = Payload(filt=filt, tuples=[tupleAction]).makePayloadEnvelopeDefer()
        d.addCallback(
            lambda payloadEnvelope: PayloadResponse(
                payloadEnvelope, destVortexName=self._destVortexName
            )
        )

        # Convert the data to TupleAction
        d.addCallback(
            lambda payloadEnvelope: payloadEnvelope.decodePayloadDefer()
        )
        d.addCallback(lambda payload: payload.tuples)
        return d



import logging
import sys
from copy import copy
from typing import Optional, Set, Union

from twisted.internet.defer import (
    succeed,
    fail,
    Deferred,
    TimeoutError,
    inlineCallbacks,
)
from twisted.internet.threads import deferToThread
from twisted.python.failure import Failure

from vortex.DeferUtil import yesMainThread
from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.PayloadPriority import RPC_PRIORITY
from vortex.PayloadResponse import PayloadResponse
from vortex.Tuple import Tuple, addTupleType, TupleField
from vortex.VortexFactory import VortexFactory

logger = logging.getLogger(__name__)


@addTupleType
class _VortexRPCArgTuple(Tuple):
    """Vortex RPC Arg Tuple

    This tuple stores the arguments used to call the remote method.

    """

    __tupleType__ = __name__ + "_VortexRPCArgTuple"

    args = TupleField(defaultValue=[])
    kwargs = TupleField(defaultValue={})


@addTupleType
class _VortexRPCResultTuple(Tuple):
    """Vortex RPC Result Tuple

    This tuple stores the result from the remote procedure call

    """

    __tupleType__ = __name__ + "_VortexRPCResultTuple"

    result = TupleField(defaultValue=None)


class _VortexRPC:
    """Vortex RPC Wrapper Class

    This wrapper class handles the mechanics of listening for the RPC calls (handler)
     and sending calls (PayloadResponse)

    Under the covers, This is what the class.

    #.  The caller will be given a Deferred
    #.  The args, kwargs for Tuples and primitive values will be serialised into payload.
    #.  The payload will be sent across the vortex
    #.  The payload will be deserialized and delivered to a PayloadEndpoint created
            for the decorated method.
    #.  The handler will then call the method.
    #.  The result from the method will then be serialised and sent back to the calling
            vortex.
    #.  The deferred will be called with the result from the remote method.

    """

    __registeredFuncNames: Set[str] = set()

    def __init__(
        self,
        func,
        listeningVortexName: str,
        timeoutSeconds: float,
        acceptOnlyFromVortex: Optional[Union[str, tuple]],
        additionalFilt: dict,
        deferToThread: bool,
        inlineCallbacks: bool,
    ) -> None:
        """

        :param listeningVortexName: If the local vortex name matches this name, then
                a handler will be setup to listen for payloads for this RPC method.

        :param timeoutSeconds: The seconds to wait for a response before calling the
                                deferreds errback with a TimeoutError

        :param acceptOnlyFromVortex: Accept payloads (calls) only from this vortex.
                The vortex can be str or tuple of str, or None to accept from any.

        :param additionalFilt: If specified, the items from this dict will be added
                                to the filt that this RPCs handler listens on.

        :param deferToThread: Should the function be called in a thread, or in the
                        reactors main loop.

        :param inlineCallbacks: Should the function be wrapped in the twisted
                @inlinecallbacks decorator before it's called?.

        """

        self.__func = func
        self.__funcSelf = None
        self.__listeningVortexName = listeningVortexName
        self.__timeoutSeconds = timeoutSeconds
        self.__acceptOnlyFromVortex = acceptOnlyFromVortex
        if isinstance(self.__acceptOnlyFromVortex, str):
            self.__acceptOnlyFromVortex = (self.__acceptOnlyFromVortex,)
        self.__deferToThread = deferToThread
        self.__inlineCallbacks = inlineCallbacks

        self.__funcName = ""
        if func.__globals__["__spec__"]:
            self.__funcName += func.__globals__["__spec__"].name
        self.__funcName += "." + func.__qualname__

        if self.__funcName in self.__registeredFuncNames:
            raise Exception(
                "RPC function name %s is already registered" % self.__funcName
            )
        self.__registeredFuncNames.add(self.__funcName)

        # Define the FILT
        self._filt = {"_internal": "vortexRPC", "key": self.__funcName}
        self._filt.update(additionalFilt)

        # Define the Endpoint

    def start(self, funcSelf=None):
        """Start

        If this is a class method, then bind the function to the
                            object passed in by bindToSelf.

        :param funcSelf: The object to bind the class instance methods self to.

        """
        if VortexFactory.isVortexNameLocal(self.__listeningVortexName):
            self.__ep = PayloadEndpoint(self._filt, self._processCall)
            logger.debug("RPC %s listening", self.__funcName)

        else:
            logger.error(
                "Ignoring request to start listening for RPC %s "
                "as vortex name %s is not local",
                self.__funcName,
                self.__listeningVortexName,
            )

        self.__funcSelf = funcSelf

        return self

    def shutdown(self):
        """Shutdown

        Shuts down the RPC PayloadEndpoint
        """
        self.__ep.shutdown()
        self.__func = None
        self.__funcSelf = None

    @inlineCallbacks
    def _processCall(
        self,
        payloadEnvelope: PayloadEnvelope,
        vortexName,
        sendResponse,
        *args,
        **kwargs
    ):
        """Process

        Process the incoming RPC call payloads.

        """
        # If the sending vortex, is local, then ignore it, RPC can not be called locally
        if VortexFactory.isVortexNameLocal(vortexName):
            logger.warning(
                "Received RPC call to %s, from local vortex %s, ignoring it",
                self.__funcName,
                vortexName,
            )
            return

        # Apply the "allow" logic
        if (
            self.__acceptOnlyFromVortex
            and vortexName not in self.__acceptOnlyFromVortex
        ):
            logger.debug(
                "Call from non-accepted vortex %s, allowing only from %s",
                vortexName,
                str(self.__acceptOnlyFromVortex),
            )
            return

        # Get the args tuple
        payload = yield payloadEnvelope.decodePayloadDefer()
        argsTuple = payload.tuples[0]
        assert isinstance(argsTuple, _VortexRPCArgTuple), (
            "argsTuple is not an instance of %s" % _VortexRPCArgTuple
        )

        logger.debug("Received RPC call for %s", self.__funcName)

        # Call the method and setup the callbacks
        result = yield self.callLocally(argsTuple.args, argsTuple.kwargs)
        yield self._processCallCallback(
            result, sendResponse, payloadEnvelope.filt
        )

    @inlineCallbacks
    def _processCallCallback(self, result, sendResponseCallable, filt):

        payloadEnvelope = yield (
            Payload(
                filt=filt, tuples=[_VortexRPCResultTuple(result=result)]
            ).makePayloadEnvelopeDefer()
        )

        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()

        yield sendResponseCallable(vortexMsg, RPC_PRIORITY)

    @inlineCallbacks
    def callForVortexUuid(self, vortexUuid: str, *args, **kwargs):
        yesMainThread()
        return (
            yield self.__callForVortexName(None, vortexUuid, *args, **kwargs)
        )

    @inlineCallbacks
    def __call__(self, *args, **kwargs):
        """Call"""
        yesMainThread()
        return (
            yield self.__callForVortexName(
                self.__listeningVortexName, None, *args, **kwargs
            )
        )

    @inlineCallbacks
    def __callForVortexName(
        self,
        vortexName: Optional[str] = None,
        vortexUuid: Optional[str] = None,
        *args,
        **kwargs
    ):
        assert vortexName or vortexUuid

        try:
            # FAKE Exception so we can raise a better stack trace later
            raise Exception()
        except:
            stack = sys.exc_info()[2]

        logger.debug("Calling RPC for %s", self.__funcName)

        payloadEnvelope = yield (
            Payload(
                filt=copy(self._filt),
                tuples=[_VortexRPCArgTuple(args=args, kwargs=kwargs)],
            ).makePayloadEnvelopeDefer(compressionLevel=4)
        )

        pr = PayloadResponse(
            payloadEnvelope,
            timeout=self.__timeoutSeconds,
            resultCheck=False,
            logTimeoutError=False,
            destVortexName=vortexName,
            destVortexUuid=vortexUuid,
        )

        # Delete the payload, we don't need to keep it in memory while we
        # get the result.
        del payloadEnvelope

        pr.addCallback(self._processResponseCallback, stack)
        pr.addErrback(self._processResponseErrback, stack)

        val = yield pr
        return val

    @inlineCallbacks
    def _processResponseCallback(self, payloadEnvelope: PayloadEnvelope, stack):
        """Process Response Callback

        Convert the PayloadResponse payload to the result from the remotely called
        method.

        """

        if not payloadEnvelope.result in (None, True):
            return Failure(
                Exception(payloadEnvelope.result).with_traceback(stack),
                exc_tb=stack,
            )

        # Get the Result from the payload
        payload = yield payloadEnvelope.decodePayloadDefer()
        resultTuple = payload.tuples[0]
        assert isinstance(resultTuple, _VortexRPCResultTuple), (
            "resultTuple is not an instance of %s" % _VortexRPCResultTuple
        )

        logger.debug("Received RPC result for %s", self.__funcName)

        # Return the remote result
        return resultTuple.result

    def _processResponseErrback(self, failure, stack):
        """Process Response Errback

        Convert the PayloadResponse payload to the result from the remotely called
        method.

        """

        if failure.check(TimeoutError):
            logger.error("Received RPC timeout for %s", self.__funcName)

            return Failure(
                Exception(
                    "RPC call timed out for %s", self.__funcName
                ).with_traceback(stack),
                exc_tb=stack,
            )

        return failure

    def callLocally(self, args, kwargs):
        """Call Locally

        This method calls the wrapped function locally, ensuring it returns a
        deferred as it's result.

        """
        try:
            if self.__funcSelf:
                args = [self.__funcSelf] + args

            if self.__inlineCallbacks:
                result = inlineCallbacks(self.__func)(*args, **kwargs)

            elif self.__deferToThread:
                result = deferToThread(self.__func, *args, **kwargs)

            else:
                result = self.__func(*args, **kwargs)

        except Exception as e:
            return fail(Failure(e))

        if isinstance(result, Deferred):
            return result

        elif isinstance(result, Failure):
            return fail(result)

        else:
            return succeed(result)


def vortexRPC(
    listeningVortexName: str,
    timeoutSeconds: float = 30.0,
    acceptOnlyFromVortex: Optional[Union[str, tuple]] = None,
    additionalFilt: Optional[dict] = None,
    deferToThread: bool = False,
    inlineCallbacks: bool = False,
):
    """Vortex RPC Decorator

    :param listeningVortexName: If the local vortex name matches this name, then
            a handler will be setup to listen for payloads for this RPC method.

    :param timeoutSeconds: The seconds to wait for a response before calling the
                            deferreds errback with a TimeoutError

    :param acceptOnlyFromVortex: Accept payloads (calls) only from this vortex.
            Or None to accept from any.

    :param additionalFilt: If specified, the items from this dict will be added
                            to the filt that this RPCs handler listens on.

    :param deferToThread: Should the function be called in a thread, or in the
                        reactors main loop.

    :param inlineCallbacks: Should the function be wrapped in the twisted
            @inlinecallbacks decorator before it's called?.

    :return A wrapped function, that will now work as an RPC call.


    EXAMPLE:

    Declare the function and decorate it.
    Because this will be imported on both sides of the vortex, we specify the name
    so that it will only be listening on one side ::


            @vortexRPC("listenVortexName")
            def myRemoteAddMethod(arg1:int, kwarg1:int=0) -> int:
                return arg1 + kwarg1


    Call the method

    .. note:: The method will ALWAYS return a deferred, regardless of if the method
                wrapped returns one or not.

    ::


            deferred = myRemoteAddMethod(5, kwarg1=7)
            print(deferred.addCallback(lambda v:print(v)))



    """
    additionalFilt = additionalFilt if additionalFilt else {}

    def decorator(func):
        return _VortexRPC(
            func,
            listeningVortexName,
            timeoutSeconds,
            acceptOnlyFromVortex,
            additionalFilt,
            deferToThread,
            inlineCallbacks,
        )

    return decorator



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import tracemalloc

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks, Deferred
from twisted.trial import unittest

from vortex.test.VortexTcpMemoryLeakPayloadTest import (
    MemoryCheckerTestMixin,
    VortexTcpConnectTestMixin,
)


class VortexTcpMemoryLeakRawTest(
    unittest.TestCase, MemoryCheckerTestMixin, VortexTcpConnectTestMixin
):
    # Increase the timeout to 15 minutes
    timeout = 30 * 60

    _patchedProcessData_dataReceivedLen = 0

    def _processData_patch(self):
        VortexTcpMemoryLeakRawTest._patchedProcessData_dataReceivedLen += len(
            self._data
        )

        # Just drop the data and continue
        self._data = b""

    @inlineCallbacks
    def test_1vortexSend50mb_oneway_msg1mb_client_drop_data(self):
        tracemalloc.start(5)
        yield self._connect()

        self._memMark()

        from vortex.VortexPayloadProtocol import VortexPayloadProtocol

        VortexPayloadProtocol._processData = (
            VortexTcpMemoryLeakRawTest._processData_patch
        )

        dataSent = 0
        for _ in range(1024):
            dataSent += 1024 ** 2
            self._vortexServer.sendVortexMsg(b"x" * 1024 ** 2)

        # while dataSent < self._patchedProcessData_dataReceivedLen:
        for _ in range(2):
            print(
                "Sleeping 2 seconds, send %s vs received %s"
                % (dataSent, self._patchedProcessData_dataReceivedLen)
            )

            d = Deferred()
            reactor.callLater(2, d.callback, True)
            yield d
            self._memPrintIncrease()

        snapshot = tracemalloc.take_snapshot()
        top_stats = snapshot.statistics("lineno")

        print("[ Top 10 ]")
        for stat in top_stats[:10]:
            print(stat)

        tracemalloc.stop()

        # while dataSent < self._patchedProcessData_dataReceivedLen:
        for _ in range(2):
            print("Sleeping 2 seconds")

            d = Deferred()
            reactor.callLater(2, d.callback, True)
            yield d
            self._memPrintIncrease()

        self._memPrintIncrease()

        self._disconnect()



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
import uuid
from datetime import datetime
from http.cookiejar import CookieJar
from typing import Union, Optional, List
from urllib.parse import urlencode

import pytz
from twisted.internet import reactor, task
from twisted.internet.defer import succeed, Deferred, inlineCallbacks
from twisted.python.failure import Failure
from twisted.web.client import Agent, CookieAgent
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer
from zope.interface.declarations import implementer

from vortex.DeferUtil import isMainThread
from vortex.PayloadEnvelope import PayloadEnvelope, VortexMsgList
from vortex.PayloadPriority import DEFAULT_PRIORITY
from vortex.VortexABC import VortexABC, VortexInfo
from vortex.VortexPayloadProtocol import VortexPayloadProtocol

logger = logging.getLogger(name=__name__)


@implementer(IBodyProducer)
class _VortexClientPayloadProducer(object):
    def __init__(self, vortexMsgs) -> None:
        self.vortexMsgs = b""

        for vortexMsg in vortexMsgs:
            self.vortexMsgs += vortexMsg + b"."

        self.length = len(self.vortexMsgs)

    def startProducing(self, consumer):
        consumer.write(self.vortexMsgs)
        return succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


class VortexPayloadHttpClientProtocol(VortexPayloadProtocol):
    def __init__(self, logger, vortexClient=None):
        VortexPayloadProtocol.__init__(self, logger)
        self._vortexClient = vortexClient

    def _beat(self):
        if self._vortexClient:
            self._vortexClient._beat()

    def _nameAndUuidReceived(self, name, uuid):
        if self._vortexClient:
            self._vortexClient._setNameAndUuid(
                name=self._serverVortexName, uuid=self._serverVortexUuid
            )


class VortexClientHttp(VortexABC):
    """VortexServer Client
    Connects to a votex server
    """

    RETRY_DELAY = 1.5  # Seconds

    # The time it takes after recieving a response from the server to receive the
    INFO_PAYLOAD_TIMEOUT = 5  # Seconds

    def __init__(self, name: str) -> None:
        self._vortexName = name
        self._vortexUuid = str(uuid.uuid1())

        self._server = None
        self._port = None

        self._retrying = False

        self._serverVortexUuid = None
        self._serverVortexName = None

        self._cookieJar = CookieJar()

        self._beatTime = None
        self._beatTimeout = 15.0  # Server beats at 5 seconds

        # Start our heart beat checker
        self._beatLoopingCall = task.LoopingCall(self._checkBeat)

        self._reconnectVortexMsgs = [PayloadEnvelope().toVortexMsg()]

        self.__protocol = None

    @property
    def localVortexInfo(self) -> VortexInfo:
        return VortexInfo(name=self._vortexName, uuid=self._vortexUuid)

    @property
    def remoteVortexInfo(self) -> List[VortexInfo]:
        if not self.__protocol:
            return []

        if not self._serverVortexUuid:
            return []

        return [
            VortexInfo(name=self._serverVortexName, uuid=self._serverVortexUuid)
        ]

    @property
    def name(self):
        return self._vortexName

    def connect(self, server, port):
        self._server = server
        self._port = port

        if self._serverVortexName:
            raise Exception("Reconnecting is not implemented")

        self._beat()
        self._beatLoopingCall.start(5.0)

        deferred = Deferred()

        def checkUuid():
            if self._serverVortexName:
                deferred.callback(True)
            else:
                reactor.callLater(0.1, checkUuid)

        checkUuid()
        self.sendVortexMsg()

        return deferred

    def disconnect(self):
        if self._beatLoopingCall:
            if self._beatLoopingCall.running:
                self._beatLoopingCall.stop()
            self._beatLoopingCall = None

        self.__protocol.transport.loseConnection()

    def addReconnectVortexMsg(self, vortexMsg: bytes):
        """Add Reconnect Payload
        :param vortexMsg: An encoded PayloadEnvelope to send when the connection reconnects
        :return:
        """
        self._reconnectVortexMsgs.append(vortexMsg)

    def sendVortexMsg(
        self,
        vortexMsgs: Union[VortexMsgList, bytes, None] = None,
        vortexUuid: Optional[str] = None,
        priority: int = DEFAULT_PRIORITY,
    ) -> Deferred:
        """Send Vortex Msg

        NOTE: Priority ins't supported as there is no buffer for this class.

        """

        if vortexMsgs is None:
            vortexMsgs = self._reconnectVortexMsgs

        if not isinstance(vortexMsgs, list):
            vortexMsgs = [vortexMsgs]

        # Check if the vortexUuid matches the destination uuid
        #
        # if vortexUuid and vortexUuid != self._
        #
        # if not self.__protocol.serverVortexUuid:
        #     return []

        if isMainThread():
            return self._sendVortexMsgLater(vortexMsgs)

        return task.deferLater(reactor, 0, self._sendVortexMsgLater, vortexMsgs)

    @inlineCallbacks
    def _sendVortexMsgLater(self, vortexMsgs: VortexMsgList):
        yield None
        assert self._server
        assert vortexMsgs

        # This transport requires base64 encoding
        for index, vortexMsg in enumerate(vortexMsgs):
            if vortexMsg.startswith(b"{"):
                vortexMsgs[index] = yield PayloadEnvelope.base64EncodeDefer(
                    vortexMsg
                )

        def ebSendAgain(failure):
            self._retrying = True
            logger.debug(
                "Retrying send of %s messages : %s",
                len(vortexMsgs),
                failure.value,
            )

            return task.deferLater(
                reactor, self.RETRY_DELAY, self._sendVortexMsgLater, vortexMsgs
            )

        def cbRequest(response):
            if response.code != 200:
                msg = "Connection to vortex %s:%s failed" % (
                    self._server,
                    self._port,
                )
                logger.error(msg)
                return Failure(Exception(msg))

            elif self._retrying:
                logger.info(
                    "VortexServer client %s:%s reconnected",
                    self._server,
                    self._port,
                )

            self._retrying = False
            self.__protocol = VortexPayloadHttpClientProtocol(
                logger, vortexClient=self
            )
            response.deliverBody(self.__protocol)
            return True

        bodyProducer = _VortexClientPayloadProducer(vortexMsgs)

        agent = CookieAgent(Agent(reactor), self._cookieJar)

        args = {"vortexUuid": self._vortexUuid, "vortexName": self._vortexName}

        uri = (
            "http://%s:%s/vortex?%s"
            % (self._server, self._port, urlencode(args))
        ).encode("UTF-8")

        d = agent.request(
            b"POST",
            uri,
            Headers(
                {
                    b"User-Agent": [b"Synerty VortexServer Client"],
                    b"Content-Type": [b"text/plain"],
                }
            ),
            bodyProducer,
        )

        d.addCallback(cbRequest)
        d.addErrback(ebSendAgain)  # Must be after cbRequest
        return d

    def _beat(self):
        """Beat, Called by protocol"""
        self._beatTime = datetime.now(pytz.utc)

    def _setNameAndUuid(self, name, uuid):
        """Set Name And Uuid, Called by protocol"""
        self._serverVortexName = name
        self._serverVortexUuid = uuid

    def _checkBeat(self):
        if (
            not (datetime.now(pytz.utc) - self._beatTime).seconds
            > self._beatTimeout
        ):
            return

        if self._retrying:
            return

        self._retrying = True

        logger.info(
            "VortexServer client dead, reconnecting %s:%s"
            % (self._server, self._port)
        )

        d = self.sendVortexMsg()

        # Add a errback that handles the failure.
        d.addErrback(lambda _: None)



import logging
from copy import copy
from typing import Optional, Union

from twisted.internet.defer import TimeoutError, inlineCallbacks, Deferred
from twisted.python.failure import Failure

from vortex.DeferUtil import vortexLogFailure
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.PayloadResponse import PayloadResponse
from vortex.VortexABC import SendVortexMsgResponseCallable
from vortex.VortexFactory import VortexFactory
from vortex.handler.TupleActionProcessor import (
    TupleActionProcessor,
    TupleActionProcessorDelegateABC,
)

logger = logging.getLogger(__name__)


class TupleActionProcessorProxy:
    """Tuple Data Observable Proxy Handler

    This class proxies the TupleActions onto another destination, giving the ability
    to pass through one services into another. EG, from a client facing python service
    to a server backend.
    """

    def __init__(
        self,
        tupleActionProcessorName,
        proxyToVortexName: str,
        additionalFilt=None,
        acceptOnlyFromVortex: Optional[Union[str, tuple]] = None,
    ) -> None:
        """Constructor

        :param tupleActionProcessorName: The name of this and the other action handler
        :param proxyToVortexName: The vortex dest name to proxy requests to
        :param additionalFilt: Any additional filter keys that are required
        """
        self._proxyToVortexName = proxyToVortexName
        self._delegateProcessor = TupleActionProcessor(
            tupleActionProcessorName=tupleActionProcessorName,
            usedForProxy__=True,
        )

        self._filt = dict(
            name=tupleActionProcessorName, key="tupleActionProcessorName"
        )

        if additionalFilt:
            self._filt.update(additionalFilt)

        self._endpoint = PayloadEndpoint(
            self._filt, self._process, acceptOnlyFromVortex=acceptOnlyFromVortex
        )

    def shutdown(self):
        self._endpoint.shutdown()

    def setDelegate(
        self, tupleName: str, processor: TupleActionProcessorDelegateABC
    ):
        """Add Tuple Action Processor Delegate

        :param tupleName: The tuple name to process actions for.
        :param processor: The processor to use for processing this tuple name.

        """
        self._delegateProcessor.setDelegate(tupleName, processor)

    @inlineCallbacks
    def _process(
        self,
        payloadEnvelope: PayloadEnvelope,
        vortexName: str,
        sendResponse: SendVortexMsgResponseCallable,
        **kwargs
    ) -> None:

        # Ignore responses from the backend, these are handled by PayloadResponse
        if vortexName == self._proxyToVortexName:
            return

        # Shortcut the logic, so that we don't decode the payload unless we need to.
        if not self._delegateProcessor.delegateCount:
            yield self._processForProxy(
                payloadEnvelope, vortexName, sendResponse
            )
            return

        # If we have local processors, then work out if this tupleAction is meant for
        # the local processor.
        payload = yield payloadEnvelope.decodePayloadDefer()

        assert (
            len(payload.tuples) == 1
        ), "TupleActionProcessor:%s Expected 1 tuples, received %s" % (
            self._tupleActionProcessorName,
            len(payload.tuples),
        )

        tupleAction = payload.tuples[0]

        if self._delegateProcessor.hasDelegate(tupleAction.tupleName()):
            self._delegateProcessor._processTupleAction(
                payloadEnvelope.filt, sendResponse, tupleAction
            )
            return

        # Else, Just send it on to the delegate we're proxying for (the backend)
        yield self._processForProxy(payloadEnvelope, vortexName, sendResponse)

    @inlineCallbacks
    def _processForProxy(
        self,
        payloadEnvelope: PayloadEnvelope,
        vortexName: str,
        sendResponse: SendVortexMsgResponseCallable,
        **kwargs
    ):
        # Keep a copy of the incoming filt, in case they are using PayloadResponse
        responseFilt = copy(payloadEnvelope.filt)

        # Track the response, log an error if it fails
        # 5 Seconds is long enough.
        # VortexJS defaults to 10s, so we have some room for round trip time.
        pr = PayloadResponse(
            payloadEnvelope,
            timeout=PayloadResponse.TIMEOUT - 5,  # 5 seconds less
            resultCheck=False,
            logTimeoutError=False,
        )

        # This is not a lambda, so that it can have a breakpoint
        def reply(payloadEnvelope: PayloadEnvelope):
            payloadEnvelope.filt = responseFilt
            d: Deferred = payloadEnvelope.toVortexMsgDefer()
            d.addCallback(sendResponse)
            return d

        pr.addCallback(reply)

        pr.addCallback(
            lambda _: logger.debug("Received action response from server")
        )
        pr.addErrback(self.__handlePrFailure, payloadEnvelope, sendResponse)

        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()
        try:
            yield VortexFactory.sendVortexMsg(
                vortexMsgs=vortexMsg, destVortexName=self._proxyToVortexName
            )
        except Exception as e:
            logger.exception(e)

    @inlineCallbacks
    def __handlePrFailure(
        self,
        f: Failure,
        payloadEnvelope: PayloadEnvelope,
        sendResponse: SendVortexMsgResponseCallable,
    ):
        payload = yield payloadEnvelope.decodePayloadDefer()
        action = payload.tuples[0]
        if f.check(TimeoutError):
            logger.error(
                "Received no response from\nprocessor %s\naction %s",
                self._filt,
                action,
            )
        else:
            logger.error(
                "Unexpected error, %s\nprocessor %s\naction %s",
                f,
                self._filt,
                action,
            )

        vortexLogFailure(f, logger)

        vortexMsg = yield PayloadEnvelope(
            filt=payloadEnvelope.filt, result=str(f.value)
        ).toVortexMsgDefer()

        sendResponse(vortexMsg)



import gzip
import logging
import struct
from datetime import datetime
from io import BytesIO
from typing import Union, List, Type

import pytz
from twisted.internet import reactor
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.internet.defer import succeed
from twisted.internet.protocol import Protocol
from twisted.web._newclient import ResponseDone
from twisted.web.client import Agent, GzipDecoder, ContentDecoderAgent
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer, UNKNOWN_LENGTH
from zope.interface import implementer

from vortex.Tuple import Tuple
from vortex.restful.RestfulResource import HTTP_REQUEST

logger = logging.getLogger(__name__)


class GzippedDataHttpClient:
    """A Restful http client that sends out a binary payload"""

    DEFAULT_HEADERS = {
        "User-Agent": ["vortex/1.0"],
        "Content-Type": ["application/json"],
    }

    def __init__(
        self,
        url: Union[str, bytes],
        payload: bytes,
        headers: dict[str, List[str]],
        method: HTTP_REQUEST,
        meta: Type[Tuple],
        isPayloadGzipped: bool = True,
        compressed: bool = True,
        timeout: float = 10,
    ):
        self._httpMethod = method.value.encode()
        self._url = url.encode() if isinstance(url, str) else url
        self._payload = payload
        self._headers = {**self.DEFAULT_HEADERS, **headers}
        self._meta = meta
        self._compressed = (
            compressed  # send http requests whether gzip-compressed or not
        )
        self._isPayloadGzipped = isPayloadGzipped
        self._timeout = timeout  # connect timeout

    @inlineCallbacks
    def run(self) -> Type[Tuple]:
        agent = Agent(reactor, connectTimeout=self._timeout)
        # Add the gzip decoder
        if self._compressed:
            agent = ContentDecoderAgent(agent, [(b"gzip", GzipDecoder)])

        binaryPayloadRequestProducer = _BinaryPayloadRequestProducer(
            self._payload,
            self._meta,
            self._isPayloadGzipped,
        )
        # Make the web request
        try:
            response = yield agent.request(
                self._httpMethod,
                self._url,
                Headers(self._headers),
                binaryPayloadRequestProducer,
            )
            self._meta = binaryPayloadRequestProducer.meta
        except Exception as e:
            self._meta.exceptions.append(e)
        finally:
            if self._meta.exceptions:
                return self._meta

        self._meta.code = response.code
        self._meta.version = response.version
        self._meta.headers = {
            k.decode(): v[0].decode()
            for k, v in response.headers.getAllRawHeaders()
        }

        self._meta = yield self._cbResponse(response, self._meta)
        return self._meta

    def _cbResponse(self, response, meta):
        responseProducer = _ResponseProducer(meta)
        response.deliverBody(responseProducer)
        return responseProducer.asyncMeta


@implementer(IBodyProducer)
class _BinaryPayloadRequestProducer:
    def __init__(
        self,
        payload: bytes,
        meta: Type[Tuple],
        isPayloadGzipped: bool = True,
    ):
        self._meta = meta
        self._payload = payload
        self._isGzippedAlready = isPayloadGzipped
        self.length = self._getGzipUncompressedSize()
        self._gzipFile = None

    @property
    def meta(self):
        return self._meta

    def startProducing(self, consumer):
        self._meta.requestDate = datetime.now(tz=pytz.utc)

        payload = self._payload
        if self._isGzippedAlready:
            payload = gzip.decompress(self._payload)

        try:
            consumer.write(payload)
        except Exception as e:
            raise e

        return succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass

    def _getGzipUncompressedSize(self) -> int:
        # Uncompressed size is stored in the last 4 bytes of the gzip file.
        # We can read the binary data and convert it to an int.
        # (This will only work for files under 4GB)
        if not self._isGzippedAlready:
            return len(self._payload)

        with BytesIO(self._payload) as f:
            try:
                return struct.unpack("I", self._payload[-4:])[0]
            except Exception as e:
                # may be invalid format of gzip
                logger.exception(e)
                return UNKNOWN_LENGTH


class _ResponseProducer(Protocol):
    def __init__(self, meta: Type[Tuple]):
        self._finishedDeferred = Deferred()
        self._writeSize = 0
        self._body = b""
        self._meta = meta

    @property
    def asyncMeta(self):
        return self._finishedDeferred

    def dataReceived(self, data: bytes):
        self._body += data

    def connectionLost(self, reason):
        if isinstance(reason.value, ResponseDone):
            self._meta.responseDate = datetime.now(tz=pytz.utc)
            self._meta.body = self._body
            self._finishedDeferred.callback(self._meta)



import json
import logging
from typing import Union, Type, runtime_checkable

from twisted.internet import reactor
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.internet.protocol import Protocol
from twisted.web._newclient import ResponseDone
from twisted.web.client import Agent, GzipDecoder, ContentDecoderAgent
from twisted.web.http_headers import Headers
from zope.interface import implementer
from twisted.internet.defer import succeed
from twisted.web.iweb import IBodyProducer

from vortex.Tuple import Tuple
from vortex.restful.RestfulResource import HTTP_REQUEST

logger = logging.getLogger(__name__)


class RestfulHttpClient:
    def __init__(
        self,
        url: Union[str, bytes],
        method: HTTP_REQUEST,
        postTuple: Tuple,
        ResponseTuple: Type[Tuple],
    ):
        self._httpMethod = method.value.encode()
        self._url = url.encode() if isinstance(url, str) else url
        self._postTuple = postTuple
        self._ResponseTuple = ResponseTuple

    @inlineCallbacks
    def run(self):
        # convert tuple to json
        body = self._postTuple.tupleToRestfulJsonDict()
        body = json.dumps(body).encode("utf-8")

        # add http headers
        headers = Headers(
            {
                "User-Agent": ["synerty/1.0"],
                "Content-Type": ["application/json"],
            }
        )

        # Add the gzip decoder
        agent = ContentDecoderAgent(Agent(reactor), [(b"gzip", GzipDecoder)])

        # Make the web request
        response = yield agent.request(
            self._httpMethod, self._url, headers, _BytesProducer(body)
        )

        # Get the response data
        responseData = yield self._cbResponse(response)

        # Convert the bytes into a tuple and return
        return self._parseTuple(responseData)

    def _cbResponse(self, response):
        bodyDownloader = _RestfulBody()
        response.deliverBody(bodyDownloader)
        return bodyDownloader.asyncData

    def _parseTuple(self, bytes_):
        json_ = json.loads(bytes_.decode("utf-8"))
        return Tuple.restfulJsonDictToTupleWithValidation(
            json_, self._ResponseTuple
        )


@implementer(IBodyProducer)
class _BytesProducer:
    def __init__(self, body):
        self.body = body
        self.length = len(body)

    def startProducing(self, consumer):
        consumer.write(self.body)
        return succeed(None)

    def pauseProducing(self):
        pass

    def stopProducing(self):
        pass


class _RestfulBody(Protocol):
    def __init__(self):
        self._finishedDeferred = Deferred()
        self._writeSize = 0
        self._body = b""

    @property
    def asyncData(self):
        return self._finishedDeferred

    def dataReceived(self, data: bytes):
        logger.debug(bytes)
        self._body += data

    def connectionLost(self, reason):
        if isinstance(reason.value, ResponseDone):
            self._finishedDeferred.callback(self._body)



"""
 * Created by Synerty Pty Ltd
 *
 * This software is open source, the MIT license applies.
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
"""
import logging
from typing import Union

from twisted.internet.defer import succeed
from twisted.internet.protocol import connectionDone, Factory

from vortex.VortexPayloadProtocol import VortexPayloadProtocol
from vortex.VortexServerConnection import VortexServerConnection
from .PayloadEnvelope import VortexMsgList, PayloadEnvelope
from .PayloadPriority import DEFAULT_PRIORITY
from .VortexServer import VortexServer

logger = logging.getLogger(name=__name__)


class VortexTcpServerProtocol(VortexPayloadProtocol):
    def __init__(self, vortex: VortexServer, addr) -> None:
        VortexPayloadProtocol.__init__(self, logger)
        self._vortex = vortex
        self._addr = addr

        self._conn = None

        self._dataBuffer = b""
        self._remoteVortexUuid = None
        self._remoteVortexName = None
        self._httpSession = None

    def _beat(self):
        if self._conn:
            self._conn.beatReceived()

    def _nameAndUuidReceived(self, name, uuid):
        self._remoteVortexUuid = uuid
        self._remoteVortexName = name
        self._conn = VortexServerConnection(
            self._vortex,
            self._remoteVortexUuid,
            self._remoteVortexName,
            self._httpSession,
            self.transport,
            self._addr,
        )

        # Send a heart beat down the new connection, tell it who we are.
        connectPayloadFilt = {}
        connectPayloadFilt[PayloadEnvelope.vortexUuidKey] = self._vortex.uuid()
        connectPayloadFilt[PayloadEnvelope.vortexNameKey] = self._vortex.name()
        self._conn.write(PayloadEnvelope(filt=connectPayloadFilt).toVortexMsg())

        self._vortex.connectionOpened(self._httpSession, self._conn)

    def _createResponseSenderCallable(self):
        def sendResponse(
            vortexMsgs: Union[VortexMsgList, bytes],
            priority: int = DEFAULT_PRIORITY,
        ):
            if isinstance(vortexMsgs, bytes):
                vortexMsgs = [vortexMsgs]

            for vortexMsg in vortexMsgs:
                self._conn.write(vortexMsg, priority)

            return succeed(True)

        return sendResponse

    def dataReceived(self, data):

        if self._vortex.isShutdown():
            return None

        VortexPayloadProtocol.dataReceived(self, data)

    def connectionLost(self, reason=connectionDone):
        if self._conn:
            self._conn.transportClosed()

        VortexPayloadProtocol.connectionLost(self, reason)


class VortexTcpServerFactory(Factory):
    protocol = None

    def __init__(self, vortexServer: VortexServer) -> None:
        self._vortexServer = vortexServer

    def buildProtocol(self, addr):
        p = VortexTcpServerProtocol(self._vortexServer, addr)
        p.factory = self
        return p



import logging
from abc import abstractmethod, ABCMeta
from typing import Dict, Optional, Union
from inspect import signature

from twisted.internet.defer import Deferred, fail, succeed, inlineCallbacks
from twisted.python import failure
from twisted.python.failure import Failure

from vortex.DeferUtil import vortexLogFailure
from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.TupleAction import TupleActionABC
from vortex.VortexABC import SendVortexMsgResponseCallable
from vortex.TupleActionVortex import TupleActionVortex

logger = logging.getLogger(__name__)


class TupleActionProcessorDelegateABC(metaclass=ABCMeta):
    @abstractmethod
    def processTupleAction(self, tupleAction: TupleActionABC,
                           **kwargs) -> Deferred:
        """Process Tuple Action

        The method generates the vortexMsg for the vortex to send.

        :param tupleAction: The C{TupleAction} to process.

        """


class TupleActionProcessor:
    def __init__(
            self,
            tupleActionProcessorName: str,
            additionalFilt: Optional[Dict] = None,
            defaultDelegate: Optional[TupleActionProcessorDelegateABC] = None,
            acceptOnlyFromVortex: Optional[Union[str, tuple]] = None,
            usedForProxy__=False,
    ) -> None:
        """Constructor

        :param tupleActionProcessorName: The name of this observable

        :param additionalFilt: Any additional filter keys that are required

        :param defaultDelegate: The default delegate to send all actions to

        :param acceptOnlyFromVortex: Accept requests only from this vortex,
            The vortex can be str or tuple of str, or None to accept from any.
        """

        self._tupleActionProcessorName = tupleActionProcessorName
        self._defaultDelegate = defaultDelegate

        self._tupleProcessorsByTupleName: Dict[
            str, TupleActionProcessorDelegateABC
        ] = {}

        if not usedForProxy__:
            self._filt = dict(
                name=tupleActionProcessorName, key="tupleActionProcessorName"
            )

            if additionalFilt:
                self._filt.update(additionalFilt)

            self._endpoint = PayloadEndpoint(
                self._filt,
                self._process,
                acceptOnlyFromVortex=acceptOnlyFromVortex,
            )

    def setDelegate(
            self, tupleName: str, processor: TupleActionProcessorDelegateABC
    ):
        """Add Tuple Action Processor Delegate

        :param tupleName: The tuple name to process actions for.
        :param processor: The processor to use for processing this tuple name.

        """
        assert (
            not tupleName in self._tupleProcessorsByTupleName
        ), "TupleActionProcessor:%s, Tuple name %s is already registered" % (
            self._tupleActionProcessorName,
            tupleName,
        )

        assert isinstance(processor, TupleActionProcessorDelegateABC), (
                "TupleActionProcessor:%s, provider must be an"
                " instance of TupleActionProcessorDelegateABC"
                % self._tupleActionProcessorName
        )

        self._tupleProcessorsByTupleName[tupleName] = processor

    @property
    def delegateCount(self) -> int:
        return len(self._tupleProcessorsByTupleName)

    def hasDelegate(self, tupleName: str) -> bool:
        return tupleName in self._tupleProcessorsByTupleName

    def shutdown(self):
        self._endpoint.shutdown()

    @inlineCallbacks
    def _process(
            self,
            payloadEnvelope: PayloadEnvelope,
            sendResponse: SendVortexMsgResponseCallable,
            **kwargs
    ):
        """Process the Payload / Tuple Action"""

        payload = yield payloadEnvelope.decodePayloadDefer()

        assert (
                len(payload.tuples) == 1
        ), "TupleActionProcessor:%s Expected 1 tuples, received %s" % (
            self._tupleActionProcessorName,
            len(payload.tuples),
        )

        tupleAction = payload.tuples[0]

        self._processTupleAction(
            payloadEnvelope.filt, sendResponse, tupleAction, **kwargs
        )

    def _processTupleAction(
            self, payloadEnvelopeFilt, sendResponse, tupleAction, **kwargs
    ):

        assert isinstance(
            tupleAction, TupleActionABC
        ), "TupleActionProcessor:%s Expected TupleAction, received %s" % (
            self._tupleActionProcessorName,
            tupleAction.__class__,
        )

        tupleName = tupleAction.tupleName()
        processor = self._tupleProcessorsByTupleName.get(tupleName)
        if processor:
            delegate = processor.processTupleAction

        elif self._defaultDelegate:
            delegate = self._defaultDelegate.processTupleAction

        else:
            raise Exception("No delegate registered for %s" % tupleName)

        sig = signature(delegate)
        if len(sig.parameters) == 2:
            tupleActionVortex = TupleActionVortex(**kwargs)
            d = self._customMaybeDeferred(delegate, tupleAction,
                                          tupleActionVortex)
        else:
            d = self._customMaybeDeferred(delegate, tupleAction)

        d.addCallback(
            self._callback, payloadEnvelopeFilt, tupleName, sendResponse
        )
        d.addErrback(
            self._errback, payloadEnvelopeFilt, tupleName, sendResponse
        )

    @inlineCallbacks
    def _callback(
            self,
            result,
            replyFilt: dict,
            tupleName: str,
            sendResponse: SendVortexMsgResponseCallable,
    ):

        if not isinstance(result, list):
            result = [result]
        payload = Payload(filt=replyFilt, tuples=result)

        payloadEnvelope = yield payload.makePayloadEnvelopeDefer()
        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()
        try:
            yield sendResponse(vortexMsg)

        except Exception as e:
            logger.error(
                "Failed to send TupleAction response for %s", tupleName
            )
            logger.exception(e)

    @inlineCallbacks
    def _errback(
            self,
            result: Failure,
            replyFilt: dict,
            tupleName: str,
            sendResponse: SendVortexMsgResponseCallable,
    ):

        logger.error(
            "TupleActionProcessor:%s Failed to process TupleActon",
            self._tupleActionProcessorName,
        )

        vortexLogFailure(result, logger)
        failureMessage = result.getErrorMessage()

        payloadEnvelope = PayloadEnvelope(filt=replyFilt, result=failureMessage)
        vortexMsg = yield payloadEnvelope.toVortexMsgDefer()
        try:
            yield sendResponse(vortexMsg)

        except Exception as e:
            logger.error(
                "Failed to send TupleAction response for %s\n%s",
                tupleName,
                failureMessage,
            )
            logger.exception(e)

    def _customMaybeDeferred(self, f, *args, **kw):
        try:
            result = f(*args, **kw)
        except Exception as e:
            return fail(failure.Failure(e))

        if isinstance(result, Deferred):
            return result
        elif isinstance(result, failure.Failure):
            return fail(result)
        else:
            return succeed(result)



import logging

from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.Tuple import tupleForTupleName
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class TuplesProviderForDB(TuplesProviderABC):
    def __init__(self, ormSessionCreatorFunc):
        self._ormSessionCreatorFunc = ormSessionCreatorFunc

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict, tupleSelector: TupleSelector) -> bytes:
        """Make VortexMsg for DB

        Considerations for this method.

        #.  It must return a vortexMsg (bytes), this ensures all the database access
                and lazy loading is completed in this thread, in this session.

        #.  It must ensure the session is closed on/after exit.

        """

        ormSession = self._ormSessionCreatorFunc()
        try:
            TupleClass = tupleForTupleName(tupleSelector.name)
            qry = ormSession.query(TupleClass)
            for key, value in tupleSelector.selector.items():
                qry = qry.filter(getattr(TupleClass, key) == value)

            return (
                Payload(filt=filt, tuples=qry.all())
                .makePayloadEnvelope()
                .toVortexMsg()
            )
        finally:
            ormSession.close()



import json
from datetime import datetime
from datetime import timedelta
from typing import Dict
from typing import Set

import pytz
from twisted.internet import reactor

from vortex.Tuple import TupleField, Tuple, addTupleType


@addTupleType
class _DebounceArgsTuple(Tuple):
    __tupleType__ = __name__ + "._DebounceArgsTuple"
    args = TupleField()
    kwargs = TupleField()


def debounceCall(debounceSeconds: float):
    """Call a debounced function that delays invoking wrapped function until after wait n seconds

    :param debounceSeconds: cooling down peroid of n seconds
    :return: the wrapped function
    """

    class Wrap:
        _lastCleanup = datetime.now(pytz.utc)
        _cleanupDelta = timedelta(minutes=1, seconds=debounceSeconds)

        _debounceTimeDelta = timedelta(seconds=debounceSeconds)

        _calls: Dict[str, datetime] = {}
        _callsQueued: Set[tuple] = set()

        def __init__(self, f):
            self._f = f

        def __cleanup(self):
            now = datetime.now(pytz.utc)
            # Cleanup the old values
            if self._cleanupDelta < now - self._lastCleanup:
                self._lastCleanup = now

                for key, val in list(self._calls.items()):
                    if (
                        self._cleanupDelta < now - val
                        and key not in self._callsQueued
                    ):
                        self._calls.pop(key)

        def __wrapArgsAndKwargs(self, *args, **kwargs):
            hashedArgs = _DebounceArgsTuple(
                args=args, kwargs=kwargs
            ).toJsonDict()
            return json.dumps(hashedArgs)

        def call(self, funcSelf, *args, **kwargs):
            now = datetime.now(pytz.utc)

            # Create a hash of the args
            hashedArgs = self.__wrapArgsAndKwargs(args, kwargs)

            # If this method has never been called before, call it
            lastCallDate = self._calls.get(hashedArgs)
            if not lastCallDate:
                self._calls[hashedArgs] = now
                return self._f(funcSelf, *args, **kwargs)

            # Work out how long it's been since this method has been called
            timeSinceLastCall = now - lastCallDate

            # If this method is overdue for a call, call it
            if timedelta(seconds=debounceSeconds) < timeSinceLastCall:
                # If this call was queued, remove it
                if hashedArgs in self._callsQueued:
                    self._callsQueued.remove(hashedArgs)
                self._calls[hashedArgs] = now
                return self._f(funcSelf, *args, **kwargs)

            timeToNextCall = (
                timedelta(seconds=debounceSeconds) - timeSinceLastCall
            )

            # If not, schedule this method for call after the debounce
            # but only if there isn't already a call queued
            if hashedArgs not in self._callsQueued:
                self._callsQueued.add(hashedArgs)
                reactor.callLater(
                    timeToNextCall.total_seconds() + 0.1,
                    self.call,
                    funcSelf,
                    *args,
                    **kwargs
                )

            return

    def wrap(f):
        wrapC = Wrap(f)

        def wrapInner(funcSelf, *args, **kwargs):
            return wrapC.call(funcSelf, *args, **kwargs)

        return wrapInner

    return wrap



import logging
import typing
from collections import defaultdict
from twisted.internet._sslverify import trustRootFromCertificates
from typing import Union, List, Optional, Dict

from OpenSSL import SSL
from rx.subjects import Subject
from twisted.internet import reactor, ssl
from twisted.internet.defer import Deferred, DeferredList, succeed
from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import deferLater
from twisted.python.failure import Failure
from txhttputil.util.PemUtil import (
    parsePemBundleForServer,
    parsePemBundleForClient,
    parsePemBundleForTrustedPeers,
)
from txhttputil.util.SslUtil import (
    parseTrustRootFromBundle,
    buildCertificateOptionsForTwisted,
)
from txwebsocket.txws import WebSocketFactory

from vortex.DeferUtil import yesMainThread
from vortex.PayloadEnvelope import VortexMsgList, PayloadEnvelope
from vortex.PayloadIO import PayloadIO
from vortex.PayloadPriority import DEFAULT_PRIORITY
from vortex.VortexABC import VortexABC
from vortex.VortexClientHttp import VortexClientHttp
from vortex.VortexClientTcp import VortexClientTcp
from vortex.VortexClientWebsocketFactory import VortexClientWebsocketFactory
from vortex.VortexServer import VortexServer
from vortex.VortexServerHttpResource import VortexServerHttpResource
from vortex.VortexServerTcp import VortexTcpServerFactory
from vortex.VortexServerWebsocket import (
    VortexWebsocketServerFactory,
    VortexWebSocketUpgradeResource,
    VortexWrappedWebSocketFactory,
)

logger = logging.getLogger(__name__)

broadcast = None

browserVortexName = "browser"


class NoVortexException(Exception):
    """No Vortex Exception

    This is raised when a remote vortex doesn't exist.

    """


VortexUuidList = List[str]


class VortexFactory:
    __vortexServersByName: Dict[str, List[VortexABC]] = defaultdict(list)
    __vortexClientsByName: Dict[str, List[VortexABC]] = defaultdict(list)

    __vortexStatusChangeSubjectsByName: Dict[str, Subject] = {}

    __isShutdown = False

    __listeningPorts = []

    def __init__(self):
        raise Exception("Vortex Factory should not be instantiated")

    @classmethod
    @inlineCallbacks
    def shutdown(cls):
        cls.__isShutdown = True
        for vortex in VortexFactory._allVortexes():
            if hasattr(vortex, "close"):
                vortex.close()
                yield deferLater(reactor, 0.05, lambda: None)
        while cls.__listeningPorts:
            cls.__listeningPorts.pop().stopListening()
            yield deferLater(reactor, 0.05, lambda: None)

    @classmethod
    def _getVortexSendRefs(
        cls, name=None, uuid=None
    ) -> List[typing.Tuple[VortexABC, List[str]]]:
        assert name or uuid

        results = []

        # logger.debug("-" * 80)
        for vortex in cls._allVortexes():
            uuids: List[str] = []
            # logger.debug("FROM : %s", vortex.localVortexInfo)
            for remoteVortexInfo in vortex.remoteVortexInfo:
                # logger.debug("        REMOTE : %s", remoteVortexInfo)
                if (name is None or remoteVortexInfo.name == name) and (
                    uuid is None or remoteVortexInfo.uuid == uuid
                ):
                    uuids.append(remoteVortexInfo.uuid)

            if uuids:
                results.append((vortex, uuids))

        return results

    @classmethod
    def _allVortexes(cls):
        """All Vortexes

        :return: A list of all the vortexes, both client and server
        """
        vortexes = []
        for vortexList in cls.__vortexServersByName.values():
            vortexes += vortexList
        for vortexList in cls.__vortexClientsByName.values():
            vortexes += vortexList

        return vortexes

    @classmethod
    def createServer(cls, name: str, rootResource) -> None:
        """Create Server

        Create a vortex server, VortexServer clients connect to this vortex serer via HTTP(S)

        VortexServer clients will connect and provide their names. This allows the factory to
        abstract away the vortex UUIDs from their names.

        :param name: The name of the local vortex.
        :param rootResource: The resource to add the vortex to. An implementation of
            C{twisted.web.resource.IResource}
        :return: None
        """

        vortexServer = VortexServer(name)
        cls.__vortexServersByName[name].append(vortexServer)

        vortexResource = VortexServerHttpResource(vortexServer)
        rootResource.putChild(b"vortex", vortexResource)

    @classmethod
    def createWebsocketServer(
        cls,
        name: str,
        port: int,
        enableSsl: Optional[bool] = False,
        sslEnableMutualTLS: Optional[bool] = False,
        sslBundleFilePath: Optional[str] = None,
        sslMutualTLSCertificateAuthorityBundleFilePath: Optional[str] = None,
    ) -> None:
        """Create Server

        Create a vortex server, VortexServer clients connect to this vortex serer via HTTP(S)

        VortexServer clients will connect and provide their names. This allows the factory to
        abstract away the vortex UUIDs from their names.

        :param name: The name of the local vortex.
        :param port: The tcp port to listen on
        :param enableSsl: switch ssl on or off for HTTP
        :param sslEnableMutualTLS: switch on or off mTLS
        :param sslBundleFilePath: a filepath to a PEM file that contains
                            a private key, a certificate and a CA certificate
                            to identify the tls server itself
        :param sslMutualTLSCertificateAuthorityBundleFilePath: a filepath to a
                            PEM file that contains all CA certificates which are
                            used for mutualTLS to verify the identities of the
                            tls clients

        :return: None
        """

        vortexServer = VortexServer(name, requiresBase64Encoding=False)
        cls.__vortexServersByName[name].append(vortexServer)
        vortexWebsocketServerFactory = VortexWebsocketServerFactory(
            vortexServer
        )
        site = WebSocketFactory(vortexWebsocketServerFactory)
        if enableSsl:
            trustedCertificateAuthorities = None
            if sslEnableMutualTLS:
                trustedCertificateAuthorities = parseTrustRootFromBundle(
                    sslMutualTLSCertificateAuthorityBundleFilePath
                )
                trustedCertificateAuthorities = trustRootFromCertificates(
                    trustedCertificateAuthorities
                )

            privateKeyWithFullChain = parsePemBundleForServer(sslBundleFilePath)
            sslContextFactory = buildCertificateOptionsForTwisted(
                privateKeyWithFullChain, trustRoot=trustedCertificateAuthorities
            )
            port = reactor.listenSSL(port, site, sslContextFactory)
        else:
            port = reactor.listenTCP(port, site)
        cls.__listeningPorts.append(port)

    @classmethod
    def createHttpWebsocketResource(
        cls, name: str
    ) -> VortexWebSocketUpgradeResource:
        vortexServer = VortexServer(name)
        cls.__vortexServersByName[name].append(vortexServer)

        vortexWebsocketServerFactory = VortexWebsocketServerFactory(
            vortexServer
        )
        websocketFactory = VortexWrappedWebSocketFactory(
            vortexWebsocketServerFactory
        )

        return VortexWebSocketUpgradeResource(websocketFactory)

    @classmethod
    def createHttpWebsocketServer(cls, name: str, rootResource) -> None:
        """Create Server

        Create a vortex server, VortexServer clients connect to this vortex serer via HTTP(S)

        VortexServer clients will connect and provide their names. This allows the factory to
        abstract away the vortex UUIDs from their names.

        :param name: The name of the local vortex.
        :param port: The tcp port to listen on
        :return: None
        """

        vortexServer = VortexServer(name)
        cls.__vortexServersByName[name].append(vortexServer)
        vortexWebsocketServerFactory = VortexWebsocketServerFactory(
            vortexServer
        )
        websocketFactory = VortexWrappedWebSocketFactory(
            vortexWebsocketServerFactory
        )

        websocketResource = VortexWebSocketUpgradeResource(websocketFactory)
        rootResource.putChild(b"vortexws", websocketResource)

    @classmethod
    def createTcpServer(cls, name: str, port: int) -> None:
        """Create Server

        Create a vortex server, VortexServer clients connect to this vortex serer via HTTP(S)

        VortexServer clients will connect and provide their names. This allows the factory to
        abstract away the vortex UUIDs from their names.

        :param name: The name of the local vortex.
        :param port: The tcp port to listen on
        :return: None
        """

        vortexServer = VortexServer(name)
        cls.__vortexServersByName[name].append(vortexServer)
        vortexTcpServerFactory = VortexTcpServerFactory(vortexServer)
        port = reactor.listenTCP(port, vortexTcpServerFactory)
        cls.__listeningPorts.append(port)

    @classmethod
    def createHttpClient(cls, name: str, host: str, port: int) -> Deferred:
        """Create Client

        Connect to a vortex Server.

        :param name: The name of the local vortex.
        :param host: The hostname of the remote vortex.
        :param port: The port of the remote vortex.

        :return: A deferred from the VortexHttpClient.connect method
        """

        logger.info("Connecting to Peek Server HTTP %s:%s", host, port)

        vortexClient = VortexClientHttp(name)
        cls.__vortexClientsByName[name].append(vortexClient)

        return vortexClient.connect(host, port)

    @classmethod
    def createWebsocketClient(
        cls,
        name: str,
        host: str,
        port: str,
        url: str,
        sslEnableMutualTLS: bool = False,
        sslClientCertificateBundleFilePath: str = None,
        sslMutualTLSCertificateAuthorityBundleFilePath: str = None,
        sslMutualTLSTrustedPeerCertificateBundleFilePath: str = None,
    ) -> Deferred:
        """Create websocket client

        Connect to a vortex websocket server
        :param name: The name of the local vortex
        :param host: The hostname of the remote vortex
        :param port: The port of the remote vortex
        :param url: The websocket url that this client tries to connect to
        :param sslEnableMutualTLS: switch on or off mTLS
        :param sslClientCertificateBundleFilePath: a PEM bundle file that
                    contains a pair of key and certificate for this web client
        :param sslMutualTLSCertificateAuthorityBundleFilePath：CA bundle file path
                                            for TLS client authentication

        :param sslMutualTLSTrustedPeerCertificateBundleFilePath: a PEM bundle
                    file that contains certificates of all trusted peers. Each
                    certificate means a PEM with no intermediate CAs and/or
                    root CAs.
        :return: A deferred from the autobahn.twisted.connectWS method
        """
        logger.info("Connecting to Peek Websocket Server %s:%s", host, port)
        vortexWebsocketClientFactory = VortexClientWebsocketFactory(
            name, url=url
        )
        cls.__vortexClientsByName[name].append(vortexWebsocketClientFactory)

        if vortexWebsocketClientFactory.isSecure and sslEnableMutualTLS:
            trustedCertificateAuthorities = parseTrustRootFromBundle(
                sslMutualTLSCertificateAuthorityBundleFilePath
            )
            trustedCertificateAuthorities = trustRootFromCertificates(
                trustedCertificateAuthorities
            )

            privateKeyWithFullChain = parsePemBundleForClient(
                sslClientCertificateBundleFilePath
            )

            trustedPeerCertificates = None
            if sslMutualTLSTrustedPeerCertificateBundleFilePath is not None:
                trustedPeerCertificates = parsePemBundleForTrustedPeers(
                    sslMutualTLSTrustedPeerCertificateBundleFilePath
                )

            sslContextFactory = buildCertificateOptionsForTwisted(
                privateKeyWithFullChain,
                trustRoot=trustedCertificateAuthorities,
                trustedPeerCertificates=trustedPeerCertificates,
            )
        else:
            # use default for http or normal https
            sslContextFactory = ssl.ClientContextFactory()
            # raise minimum version to tls 1.2
            sslContextFactory.method = SSL.TLSv1_2_METHOD

        return vortexWebsocketClientFactory.connect(
            host, port, sslContextFactory
        )

    @classmethod
    def createTcpClient(cls, name: str, host: str, port: int) -> Deferred:
        """Create Client

        Connect to a vortex Server.

        :param name: The name of the local vortex.
        :param host: The hostname of the remote vortex.
        :param port: The port of the remote vortex.

        :return: A deferred from the VortexTcpClient.connect method
        """

        logger.info("Connecting to Peek Server TCP %s:%s", host, port)

        vortexClient = VortexClientTcp(name)
        cls.__vortexClientsByName[name].append(vortexClient)

        return vortexClient.connect(host, port)

    @classmethod
    def isVortexNameLocal(cls, vortexName: str) -> bool:
        for vortex in cls._allVortexes():
            if vortex.localVortexInfo.name == vortexName:
                return True

        return False

    @classmethod
    def getLocalVortexClients(cls, localVortexName: str) -> List[VortexABC]:
        vortexes: List[VortexABC] = []

        for items in cls.__vortexClientsByName.values():
            vortexes.extend(
                filter(
                    lambda x: x.localVortexInfo.name == localVortexName, items
                )
            )

        return vortexes

    @classmethod
    def getRemoteVortexUuids(cls) -> List[str]:
        remoteUuids = []

        for vortex in cls._allVortexes():
            for remoteVortexInfo in vortex.remoteVortexInfo:
                remoteUuids.append(remoteVortexInfo.uuid)

        return remoteUuids

    @classmethod
    def getRemoteVortexName(cls) -> List[str]:
        remoteNames = set()

        for vortex in cls._allVortexes():
            for remoteVortexInfo in vortex.remoteVortexInfo:
                remoteNames.add(remoteVortexInfo.name)

        return list(remoteNames)

    @classmethod
    def getRemoteVortexInfoByIp(cls, ip: str) -> str:
        # Ignore the port if any
        ip = ip.split(":")[0]

        targetUuids = []
        for servers in cls.__vortexServersByName.values():
            for server in servers:
                for uuid, conn in server.connections.items():
                    if conn.ip == ip:
                        targetUuids.append(uuid)

        if len(targetUuids) > 1:
            raise Exception("Multiple agents found running on the same server")
        return targetUuids[0]

    @classmethod
    @inlineCallbacks
    def sendVortexMsg(
        cls,
        vortexMsgs: Union[VortexMsgList, bytes],
        destVortexName: Optional[str] = broadcast,
        destVortexUuid: Optional[str] = broadcast,
    ) -> Deferred:
        """Send VortexMsg

        Sends a payload to the remote vortex.

        :param vortexMsgs: The vortex message(s) to send to the remote vortex.

        :param destVortexName: The name of the vortex to send the payload to.
            This can be null, If provided, it's used to limit the vortexes the
             message is sent to.

        :param destVortexUuid: The uuid of the vortex to send the payload to,
            This can be null, If provided, it's used to limit the vortexes the
             message is sent to.


        :return: A C{Deferred} which will callback when the message has been sent.
        """

        vortexAndUuids = cls._getVortexSendRefs(destVortexName, destVortexUuid)
        if not vortexAndUuids:
            raise NoVortexException(
                "Can not find vortexes to send message to,"
                " name=%s, uuid=%s" % (destVortexName, destVortexUuid)
            )

        if not isinstance(vortexMsgs, list):
            vortexMsgs = [vortexMsgs]

        base64VortexMsgs = vortexMsgs[:]

        # If any transports require base64 encoding, then encode them all
        if [v for v in vortexAndUuids if v[0].requiresBase64Encoding]:
            for index, vortexMsg in enumerate(base64VortexMsgs):
                if vortexMsg.startswith(b"{"):
                    base64VortexMsgs[
                        index
                    ] = yield PayloadEnvelope.base64EncodeDefer(vortexMsg)

        deferreds = []
        for vortex, uuids in vortexAndUuids:
            for uuid in uuids:
                if vortex.requiresBase64Encoding:
                    deferreds.append(
                        vortex.sendVortexMsg(base64VortexMsgs, uuid)
                    )
                else:
                    deferreds.append(vortex.sendVortexMsg(vortexMsgs, uuid))

        results = yield DeferredList(deferreds)
        return results

    @classmethod
    def sendVortexMsgLocally(
        cls,
        vortexMsgs: Union[VortexMsgList, bytes],
        priority: int = DEFAULT_PRIORITY,
    ) -> Deferred:
        """Send VortexMsg

        Sends a payload to the remote vortex.

        :param vortexMsgs: The vortex message(s) to deliver locally.

        :return: A C{Deferred} which will callback when the message has been delivered.
        """
        yesMainThread()

        vortexUuid = "local"
        vortexName = "local"
        httpSession = "local"
        sendResponse = VortexFactory.sendVortexMsgLocally

        vortexMsgs = (
            [vortexMsgs] if isinstance(vortexMsgs, bytes) else vortexMsgs
        )

        def send(payloadEnvelope: PayloadEnvelope):
            try:
                PayloadIO().process(
                    payloadEnvelope=payloadEnvelope,
                    vortexUuid=vortexUuid,
                    vortexName=vortexName,
                    httpSession=httpSession,
                    sendResponse=sendResponse,
                )
                return succeed(True)

            except Exception as e:
                return Failure(e)

        deferreds = []
        for vortexMsg in vortexMsgs:
            d = PayloadEnvelope().fromVortexMsgDefer(vortexMsg)
            d.addCallback(send)
            deferreds.append(d)

        return DeferredList(deferreds)

    @classmethod
    def subscribeToVortexStatusChange(cls, vortexName: str) -> Subject:
        """Subscribe to Vortex Status Change

        Subscribing to the returned observable/subject will provided updates of when the
        vortex goes offline, or online.

        .. warning:: This is only implemented for TCP Client vortexes.

        :param vortexName: The name of the vortex to subscribe to the status for.
                            This will be the name of the remote vortex.

        """
        if not vortexName in cls.__vortexStatusChangeSubjectsByName:
            cls.__vortexStatusChangeSubjectsByName[vortexName] = Subject()
        return cls.__vortexStatusChangeSubjectsByName[vortexName]

    @classmethod
    def _notifyOfVortexStatusChange(cls, vortexName: str, online: bool) -> None:
        if cls.__isShutdown:
            return

        logger.debug(
            "Vortex %s went %s", vortexName, ("online" if online else "offline")
        )

        if vortexName in cls.__vortexStatusChangeSubjectsByName:
            cls.__vortexStatusChangeSubjectsByName[vortexName].on_next(online)



import logging
import unittest

from vortex.Tuple import (
    Tuple,
    addTupleType,
    TupleField,
    PolymorphicTupleTypeFieldArg,
    TupleHash,
)

logger = logging.getLogger(__name__)


@addTupleType
class Parent(Tuple):
    __tupleType__ = "com.synerty.Tuples.Parent"
    __tupleArgs__ = (PolymorphicTupleTypeFieldArg("concreteType"),)

    aStr: str = TupleField()
    concreteType: str = TupleField()


@addTupleType
class Child(Parent):
    __tupleType__ = "com.synerty.Tuples.Child"

    aInt: int = TupleField()


@addTupleType
class Another(Tuple):
    __tupleType__ = "com.synerty.Tuples.Another"

    aInt: int = TupleField()
    concreteType: str = TupleField()


class PolymorphicTupleTypeFieldArgTest(unittest.TestCase):
    def testParentTupleToFromJSON(self):
        parent1 = Parent(aStr="Test", concreteType=Parent.tupleType())
        parent2 = Parent(aStr="Test", concreteType=Parent.tupleType())
        parent3 = Parent(aStr="Test 1", concreteType=Parent.tupleType())

        self.assertEqual(TupleHash(parent1), TupleHash(parent2))
        self.assertNotEqual(TupleHash(parent1), TupleHash(parent3))

    def testChildFieldsInToJsonDict(self):
        data = dict(aStr="Test", aInt=1, concreteType=Child.tupleType())
        child = Child(**data)

        self.assertEqual(data, child.tupleToRestfulJsonDict())

    def testChildTupleToFromJSON(self):
        # Construct Child through Child
        child1 = Child(aStr="Test", aInt=1, concreteType=Child.tupleType())

        # Construct Child through Parent
        child2 = Tuple.restfulJsonDictToTupleWithValidation(
            child1.tupleToRestfulJsonDict(), Parent
        )

        self.assertEqual(TupleHash(child1), TupleHash(child2))

        # Parent should not match Child
        parent1 = Parent(aStr="Test", concreteType=Parent.tupleType())
        self.assertNotEqual(TupleHash(child1), TupleHash(parent1))

        # Parent should only be able to init its children
        another1 = Another(aInt=1, concreteType=Another.tupleType())
        with self.assertRaises(TypeError):
            Tuple.restfulJsonDictToTupleWithValidation(
                another1.tupleToRestfulJsonDict(), Parent
            )



import logging

from twisted.internet import reactor

from vortex.VortexFactory import VortexFactory
from vortex.rpc.test.RPCTest import myRemoteAddMethod

logger = logging.getLogger(__name__)
logging.basicConfig()

if __name__ == "__main__":
    VortexFactory.createWebsocketServer("listenVortexName", 10101)

    reactor.callLater(0, logger.info, "RPCTest server running")
    reactor.callLater(0, myRemoteAddMethod.start)
    reactor.run()



import io
import json
import logging

from vortex.restful.RestfulResource import HTTP_REQUEST
from vortex.restful.RestfulResource import PluginRestfulResource
from twisted.trial import unittest
from twisted.web.test.requesthelper import DummyRequest
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType

logger = logging.getLogger(__name__)


@addTupleType
class DummyTuple(Tuple):
    __tupleType__ = "DummyTuple"

    id = TupleField()
    text = TupleField()


class TestHandlerController:
    def testMethod(self, tuple_: DummyTuple):
        if hasattr(tuple_, "tupleToRestfulJsonDict"):
            # vortexpy > 2.1.3
            return tuple_.tupleToRestfulJsonDict()
        else:
            return tuple_.tupleToSmallJsonDict()


class RestfulServerTest(unittest.TestCase):
    def setUp(self):
        handlerController = TestHandlerController()
        self.pluginAResource = PluginRestfulResource()
        self.pluginAResource.registerMethod(
            handlerController.testMethod,
            DummyTuple,
            b"test",
            [HTTP_REQUEST.GET, HTTP_REQUEST.POST],
        )
        self.pluginAResource.registerMethod(
            handlerController.testMethod,
            DummyTuple,
            b"test/test",
            [HTTP_REQUEST.GET, HTTP_REQUEST.POST],
        )
        self.pluginAResource.registerMethod(
            handlerController.testMethod,
            DummyTuple,
            b"test/test/test",
            [HTTP_REQUEST.GET, HTTP_REQUEST.POST],
        )

    def _dictToBytes(self, dictionary: dict) -> bytes:
        return bytes(json.dumps(dictionary), "utf-8")

    def _bytesToDict(self, bytes_: bytes) -> dict:
        string = bytes_[0].decode("utf-8")
        return json.loads(string)

    def _check(self, expected: dict, actual: dict) -> bool:
        # expected dict should be equal or a subnet of actual dict
        return expected.items() <= actual.items()

    def testValidJsonRequest(self):
        requestDict = {"id": 1, "text": "text"}

        request = DummyRequest([])
        request.content = io.BytesIO(self._dictToBytes(requestDict))

        jsonResource = self.pluginAResource.getChild(b"test", request)

        jsonResource.render(request)
        self.assertEqual(request.responseCode, 200)
        responseDict = self._bytesToDict(request.written)
        self.assertTrue(self._check(requestDict, responseDict))

    def testInvalidJsonRequest(self):
        request = DummyRequest([])
        request.content = io.BytesIO(b"##invalid json}")

        jsonResource = self.pluginAResource.getChild(b"test", request)
        jsonResource.render(request)
        self.assertEqual(request.responseCode, 500)

    def testPath(self):
        requestDict = {"id": 1, "text": "text"}

        for path in [b"test", b"test/test", b"test/test/test"]:
            request = DummyRequest([])
            request.content = io.BytesIO(self._dictToBytes(requestDict))

            jsonResource = self.pluginAResource.getChild(path, request)

            jsonResource.render(request)
            self.assertEqual(request.responseCode, 200)
            responseDict = self._bytesToDict(request.written)
            self.assertTrue(self._check(requestDict, responseDict))
